# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
This file implements the rtnetlink API that is used to read the network
configuration of the machine.
"""

import socket
import struct
import time

import scapy.utils6

from scapy.consts import BIG_ENDIAN
from scapy.config import conf
from scapy.error import log_loading
from scapy.packet import (
    Packet,
    bind_layers,
)
from scapy.utils import atol, itom

from scapy.fields import (
    ByteEnumField,
    ByteField,
    EnumField,
    Field,
    FieldLenField,
    FlagsField,
    IP6Field,
    IPField,
    LenField,
    MACField,
    MayEnd,
    MultipleTypeField,
    PacketListField,
    PadField,
    StrLenField,
    XStrLenField,
)

from scapy.arch.common import _iff_flags

# Typing imports
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

# from <linux/netlink.h> and <linux/rtnetlink.h>


# Common header


class rtmsghdr(Packet):
    fields_desc = [
        LenField("nlmsg_len", None, fmt="=L"),
        EnumField(
            "nlmsg_type",
            0,
            {
                # netlink.h
                3: "NLMSG_DONE",
                # rtnetlink.h
                16: "RTM_NEWLINK",
                17: "RTM_DELLINK",
                18: "RTM_GETLINK",
                19: "RTM_SETLINK",
                20: "RTM_NEWADDR",
                21: "RTM_DELADDR",
                22: "RTM_GETADDR",
                # 23: unused
                24: "RTM_NEWROUTE",
                25: "RTM_DELROUTE",
                26: "RTM_GETROUTE",
                # 27: unused
            },
            fmt="=H",
        ),
        FlagsField(
            "nlmsg_flags",
            0,
            16 if BIG_ENDIAN else -16,
            {
                0x01: "NLM_F_REQUEST",
                0x02: "NLM_F_MULTI",
                0x04: "NLM_F_ACK",
                0x08: "NLM_F_ECHO",
                0x10: "NLM_F_DUMP_INTR",
                0x20: "NLM_F_DUMP_FILTERED",
                # GET modifiers
                0x100: "NLM_F_ROOT",
                0x200: "NLM_F_MATCH",
                0x400: "NLM_F_ATOMIC",
            },
        ),
        Field("nlmsg_seq", 0, fmt="=L"),
        Field("nlmsg_pid", 0, fmt="=L"),
    ]

    def post_build(self, pkt: bytes, pay: bytes) -> bytes:
        pkt += pay
        if self.nlmsg_len is None:
            pkt = struct.pack("=L", len(pkt)) + pkt[4:]
        return pkt

    def extract_padding(self, s: bytes) -> Tuple[bytes, Optional[bytes]]:
        return s[: self.nlmsg_len - 16], s[self.nlmsg_len - 16 :]

    def answers(self, other: Packet) -> bool:
        return bool(other.nlmsg_seq == self.nlmsg_seq)


# DONE


class nlmsgerr_rtattr(Packet):
    fields_desc = [
        FieldLenField(
            "rta_len", None, length_of="rta_data", fmt="=H", adjust=lambda _, x: x + 4
        ),
        EnumField(
            "rta_type",
            0,
            {},
            fmt="=H",
        ),
        PadField(
            MultipleTypeField(
                [],
                StrLenField(
                    "rta_data",
                    b"",
                    length_from=lambda pkt: pkt.rta_len - 4,
                ),
            ),
            align=4,
        ),
    ]

    def default_payload_class(self, payload: bytes) -> Type[Packet]:
        return conf.padding_layer


class nlmsgerr(Packet):
    fields_desc = [
        MayEnd(Field("status", 0, fmt="=L")),
        # Pay
        PacketListField("data", [], nlmsgerr_rtattr),
    ]


bind_layers(rtmsghdr, nlmsgerr, nlmsg_type=3)


# LINK messages


class ifla_af_spec_inet_rtattr(Packet):
    fields_desc = [
        FieldLenField(
            "rta_len", None, length_of="rta_data", fmt="=H", adjust=lambda _, x: x + 4
        ),
        EnumField(
            "rta_type",
            0,
            {
                0x00: "IFLA_INET_UNSPEC",
                0x01: "IFLA_INET_CONF",
            },
            fmt="=H",
        ),
        PadField(
            MultipleTypeField(
                [],
                XStrLenField(
                    "rta_data",
                    b"",
                    length_from=lambda pkt: pkt.rta_len - 4,
                ),
            ),
            align=4,
        ),
    ]

    def default_payload_class(self, payload: bytes) -> Type[Packet]:
        return conf.padding_layer


class ifla_af_spec_inet6_rtattr(Packet):
    fields_desc = [
        FieldLenField(
            "rta_len", None, length_of="rta_data", fmt="=H", adjust=lambda _, x: x + 4
        ),
        EnumField(
            "rta_type",
            0,
            {
                0x00: "IFLA_INET6_UNSPEC",
                0x01: "IFLA_INET6_FLAGS",
                0x02: "IFLA_INET6_CONF",
                0x03: "IFLA_INET6_STATS",
                0x04: "IFLA_INET6_MCAST",
                0x05: "IFLA_INET6_CACHEINFO",
                0x06: "IFLA_INET6_ICMP6STATS",
                0x07: "IFLA_INET6_TOKEN",
                0x08: "IFLA_INET6_ADDR_GEN_MODE",
                0x09: "IFLA_INET6_RA_MTU",
            },
            fmt="=H",
        ),
        PadField(
            MultipleTypeField(
                [],
                XStrLenField(
                    "rta_data",
                    b"",
                    length_from=lambda pkt: pkt.rta_len - 4,
                ),
            ),
            align=4,
        ),
    ]

    def default_payload_class(self, payload: bytes) -> Type[Packet]:
        return conf.padding_layer


class ifla_af_spec_rtattr(Packet):
    fields_desc = [
        FieldLenField(
            "rta_len", None, length_of="rta_data", fmt="=H", adjust=lambda _, x: x + 4
        ),
        EnumField("rta_type", 0, socket.AddressFamily, fmt="=H"),
        PadField(
            MultipleTypeField(
                [
                    (
                        # AF_INET
                        PacketListField(
                            "rta_data",
                            [],
                            ifla_af_spec_inet_rtattr,
                            length_from=lambda pkt: pkt.rta_len - 4,
                        ),
                        lambda pkt: pkt.rta_type == 2,
                    ),
                    (
                        # AF_INET6
                        PacketListField(
                            "rta_data",
                            [],
                            ifla_af_spec_inet6_rtattr,
                            length_from=lambda pkt: pkt.rta_len - 4,
                        ),
                        lambda pkt: pkt.rta_type == 10,
                    ),
                ],
                XStrLenField(
                    "rta_data",
                    b"",
                    length_from=lambda pkt: pkt.rta_len - 4,
                ),
            ),
            align=4,
        ),
    ]

    def default_payload_class(self, payload: bytes) -> Type[Packet]:
        return conf.padding_layer


class ifinfomsg_rtattr(Packet):
    fields_desc = [
        FieldLenField(
            "rta_len", None, length_of="rta_data", fmt="=H", adjust=lambda _, x: x + 4
        ),
        EnumField(
            "rta_type",
            0,
            {
                0x00: "IFLA_UNSPEC",
                0x01: "IFLA_ADDRESS",
                0x02: "IFLA_BROADCAST",
                0x03: "IFLA_IFNAME",
                0x04: "IFLA_MTU",
                0x05: "IFLA_LINK",
                0x06: "IFLA_QDISC",
                0x07: "IFLA_STATS",
                0x08: "IFLA_COST",
                0x09: "IFLA_PRIORITY",
                0x0A: "IFLA_MASTER",
                0x0B: "IFLA_WIRELESS",
                0x0C: "IFLA_PROTINFO",
                0x0D: "IFLA_TXQLEN",
                0x0E: "IFLA_MAP",
                0x0F: "IFLA_WEIGHT",
                0x10: "IFLA_OPERSTATE",
                0x11: "IFLA_LINKMODE",
                0x12: "IFLA_LINKINFO",
                0x13: "IFLA_NET_NS_PID",
                0x14: "IFLA_IFALIAS",
                0x15: "IFLA_NUM_VS",
                0x16: "IFLA_VFINFO_LIST",
                0x17: "IFLA_STATS64",
                0x18: "IFLA_VF_PORTS",
                0x19: "IFLA_PORT_SELF",
                0x1A: "IFLA_AF_SPEC",
                0x1B: "IFLA_GROUP",
                0x1C: "IFLA_NET_NS_FD",
                0x1D: "IFLA_EXT_MASK",
                0x1E: "IFLA_PROMISCUITY",
                0x1F: "IFLA_NUM_TX_QUEUES",
                0x20: "IFLA_NUM_RX_QUEUES",
                0x21: "IFLA_CARRIER",
                0x22: "IFLA_PHYS_PORT_ID",
                0x23: "IFLA_CARRIER_CHANGES",
                0x24: "IFLA_PHYS_SWITCH_ID",
                0x25: "IFLA_LINK_NETNSID",
                0x26: "IFLA_PHYS_PORT_NAME",
                0x27: "IFLA_PROTO_DOWN",
                0x28: "IFLA_GSO_MAX_SEGS",
                0x29: "IFLA_GSO_MAX_SIZE",
                0x2A: "IFLA_PAD",
                0x2B: "IFLA_XDP",
                0x2C: "IFLA_EVENT",
                0x2D: "IFLA_NEW_NETNSID",
                0x2E: "IFLA_IF_NETNSID",
                0x2F: "IFLA_CARRIER_UP_COUNT",
                0x30: "IFLA_CARRIER_DOWN_COUNT",
                0x31: "IFLA_NEW_IFINDEX",
                0x32: "IFLA_MIN_MTU",
                0x33: "IFLA_MAX_MTU",
                0x34: "IFLA_PROP_LIST",
                0x35: "IFLA_ALT_IFNAME",
                0x36: "IFLA_PERM_ADDRESS",
                0x37: "IFLA_PROTO_DOWN_REASON",
                0x38: "IFLA_PARENT_DEV_NAME",
                0x39: "IFLA_PARENT_DEV_BUS_NAME",
                0x3A: "IFLA_GRO_MAX_SIZE",
                0x3B: "IFLA_TSO_MAX_SIZE",
                0x3C: "IFLA_TSO_MAX_SEGS",
                0x3D: "IFLA_ALLMULTI",
            },
            fmt="=H",
        ),
        PadField(
            MultipleTypeField(
                [
                    (
                        # IFLA_ADDRESS
                        MACField("rta_data", "00:00:00:00:00:00"),
                        lambda pkt: pkt.rta_type in [0x01, 0x36],
                    ),
                    (
                        # IFLA_IFNAME
                        StrLenField(
                            "rta_data", b"", length_from=lambda pkt: pkt.rta_len - 4
                        ),
                        lambda pkt: pkt.rta_type in [0x03],
                    ),
                    (
                        # IFLA_AF_SPEC
                        PacketListField(
                            "rta_data",
                            [],
                            ifla_af_spec_rtattr,
                            length_from=lambda pkt: pkt.rta_len - 4,
                        ),
                        lambda pkt: pkt.rta_type == 0x1A,
                    ),
                ],
                XStrLenField(
                    "rta_data",
                    b"",
                    length_from=lambda pkt: pkt.rta_len - 4,
                ),
            ),
            align=4,
        ),
    ]

    def default_payload_class(self, payload: bytes) -> Type[Packet]:
        return conf.padding_layer


class ifinfomsg(Packet):
    fields_desc = [
        ByteEnumField("ifi_family", 0, socket.AddressFamily),
        ByteField("res", 0),
        Field("ifi_type", 0, fmt="=H"),
        Field("ifi_index", 0, fmt="=i"),
        FlagsField(
            "ifi_flags",
            0,
            32 if BIG_ENDIAN else -32,
            _iff_flags,
        ),
        Field("ifi_change", 0, fmt="=I"),
        # Pay
        PacketListField("data", [], ifinfomsg_rtattr),
    ]


bind_layers(rtmsghdr, ifinfomsg, nlmsg_type=16)
bind_layers(rtmsghdr, ifinfomsg, nlmsg_type=17)
bind_layers(rtmsghdr, ifinfomsg, nlmsg_type=18)
bind_layers(rtmsghdr, ifinfomsg, nlmsg_type=19)


# ADDR messages


class ifaddrmsg_rtattr(Packet):
    fields_desc = [
        FieldLenField(
            "rta_len", None, length_of="rta_data", fmt="=H", adjust=lambda _, x: x + 4
        ),
        EnumField(
            "rta_type",
            0,
            {
                0x00: "IFA_UNSPEC",
                0x01: "IFA_ADDRESS",
                0x02: "IFA_LOCAL",
                0x03: "IFA_LABEL",
                0x04: "IFA_BROADCAST",
                0x05: "IFA_ANYCAST",
                0x06: "IFA_CACHEINFO",
                0x07: "IFA_MULTICAST",
                0x08: "IFA_FLAGS",
                0x09: "IFA_RT_PRIORITY",
                0x0A: "IFA_TARGET_NETNSID",
                0x0B: "IFA_PROTO",
            },
            fmt="=H",
        ),
        PadField(
            MultipleTypeField(
                [
                    # IFA_ADDRESS, IFA_LOCAL, IFA_BROADCAST
                    (
                        IPField("rta_data", "0.0.0.0"),
                        lambda pkt: pkt.parent
                        and pkt.parent.ifa_family == 2
                        and pkt.rta_type in [0x01, 0x02, 0x04],
                    ),
                    (
                        IP6Field("rta_data", "::"),
                        lambda pkt: pkt.parent
                        and pkt.parent.ifa_family == 10
                        and pkt.rta_type in [0x01, 0x02, 0x04],
                    ),
                    (
                        # IFA_LABEL
                        StrLenField(
                            "rta_data", b"", length_from=lambda pkt: pkt.rta_len - 4
                        ),
                        lambda pkt: pkt.rta_type in [0x03],
                    ),
                ],
                XStrLenField(
                    "rta_data",
                    b"",
                    length_from=lambda pkt: pkt.rta_len - 4,
                ),
            ),
            align=4,
        ),
    ]

    def default_payload_class(self, payload: bytes) -> Type[Packet]:
        return conf.padding_layer


class ifaddrmsg(Packet):
    fields_desc = [
        ByteEnumField("ifa_family", 0, socket.AddressFamily),
        ByteField("ifa_prefixlen", 0),
        FlagsField(
            "ifa_flags",
            0,
            -8,
            {
                0x01: "IFA_F_SECONDARY",
                0x02: "IFA_F_NODAD",
                0x04: "IFA_F_OPTIMISTIC",
                0x08: "IFA_F_DADFAILED",
                0x10: "IFA_F_HOMEADDRESS",
                0x20: "IFA_F_DEPRECATED",
                0x40: "IFA_F_TENTATIVE",
                0x80: "IFA_F_PERMANENT",
            },
        ),
        ByteField("ifa_scope", 0),
        Field("ifa_index", 0, fmt="=L"),
        # Pay
        PacketListField("data", [], ifaddrmsg_rtattr),
    ]


bind_layers(rtmsghdr, ifaddrmsg, nlmsg_type=20)
bind_layers(rtmsghdr, ifaddrmsg, nlmsg_type=21)
bind_layers(rtmsghdr, ifaddrmsg, nlmsg_type=22)


# ROUTE messages


RT_CLASS = {
    0: "RT_TABLE_UNSPEC",
    252: "RT_TABLE_COMPAT",
    253: "RT_TABLE_DEFAULT",
    254: "RT_TABLE_MAIN",
    255: "RT_TABLE_LOCAL",
}


class rtmsg_rtattr(Packet):
    fields_desc = [
        FieldLenField(
            "rta_len", None, length_of="rta_data", fmt="=H", adjust=lambda _, x: x + 4
        ),
        EnumField(
            "rta_type",
            0,
            {
                0x00: "RTA_UNSPEC",
                0x01: "RTA_DST",
                0x02: "RTS_SRC",
                0x03: "RTS_IIF",
                0x04: "RTS_OIF",
                0x05: "RTA_GATEWAY",
                0x06: "RTA_PRIORITY",
                0x07: "RTA_PREFSRC",
                0x08: "RTA_METRICS",
                0x09: "RTA_MULTIPATH",
                0x0B: "RTA_FLOW",
                0x0C: "RTA_CACHEINFO",
                0x0F: "RTA_TABLE",
                0x10: "RTA_MARK",
                0x11: "RTA_MFC_STATS",
                0x12: "RTA_VIA",
                0x13: "RTA_NEWDST",
                0x14: "RTA_PREF",
                0x15: "RTA_ENCAP_TYPE",
                0x16: "RTA_ENCAP",
                0x17: "RTA_EXPIRES",
                0x18: "RTA_PAD",
                0x19: "RTA_UID",
                0x1A: "RTA_TTL_PROPAGATE",
                0x1B: "RTA_IP_PROTO",
                0x1C: "RTA_SPORT",
                0x1D: "RTA_DPORT",
                0x1E: "RTA_NH_ID",
            },
            fmt="=H",
        ),
        PadField(
            MultipleTypeField(
                [
                    # RTA_DST, RTA_SRC, RTA_PREFSRC, RTA_GATEWAY
                    (
                        IPField("rta_data", "0.0.0.0"),
                        lambda pkt: pkt.parent
                        and pkt.parent.rtm_family == 2
                        and pkt.rta_type in [0x01, 0x02, 0x05, 0x07],
                    ),
                    (
                        IP6Field("rta_data", "::"),
                        lambda pkt: pkt.parent
                        and pkt.parent.rtm_family == 10
                        and pkt.rta_type in [0x01, 0x02, 0x05, 0x07],
                    ),
                    # RTS_OIF, RTA_PRIORITY
                    (
                        Field("rta_data", 0, fmt="=I"),
                        lambda pkt: pkt.rta_type in [0x04, 0x06, 0x10],
                    ),
                    # RTA_TABLE
                    (
                        EnumField("rta_data", 0, RT_CLASS, fmt="=I"),
                        lambda pkt: pkt.rta_type in [0x0F],
                    ),
                ],
                XStrLenField(
                    "rta_data",
                    b"",
                    length_from=lambda pkt: pkt.rta_len - 4,
                ),
            ),
            align=4,
        ),
    ]

    def default_payload_class(self, payload: bytes) -> Type[Packet]:
        return conf.padding_layer


class rtmsg(Packet):
    fields_desc = [
        ByteEnumField("rtm_family", 0, socket.AddressFamily),
        ByteField("rtm_dst_len", 0),
        ByteField("rtm_src_len", 0),
        ByteField("rtm_tos", 0),
        ByteEnumField(
            "rtm_table",
            0,
            RT_CLASS,
        ),
        ByteEnumField(
            "rtm_protocol",
            0,
            {
                0x00: "RTPROT_UNSPEC",
                0x01: "RTPROT_REDIRECT",
                0x02: "RTPROT_KERNEL",
                0x03: "RTPROT_BOOT",
                0x04: "RTPROT_STATIC",
            },
        ),
        ByteEnumField(
            "rtm_scope",
            0,
            {
                0: "RT_SCOPE_UNIVERSE",
                200: "RT_SCOPE_SITE",
                253: "RT_SCOPE_LINK",
                254: "RT_SCOPE_HOST",
                255: "RT_SCOPE_NOWHERE",
            },
        ),
        ByteEnumField(
            "rtm_type",
            0,
            {
                0x00: "RTN_UNSPEC",
                0x01: "RTN_UNICAST",
                0x02: "RTN_LOCAL",
                0x03: "RTN_BROADCAST",
                0x04: "RTN_ANYCAST",
                0x05: "RTN_MULTICAST",
                0x06: "RTN_BLACKHOLE",
                0x07: "RTN_UNREACHABLE",
                0x08: "RTN_PROHIBIT",
                0x09: "RTN_THROW",
                0x0A: "RTN_NAT",
                0x0B: "RTN_XRESOLVE",
            },
        ),
        FlagsField(
            "rtm_flags",
            0,
            32 if BIG_ENDIAN else -32,
            {
                0x100: "RTM_F_NOTIFY",
                0x200: "RTM_F_CLONED",
                0x400: "RTM_F_EQUALIZE",
                0x800: "RTM_F_PREFIX",
                0x1000: "RTM_F_LOOKUP_TABLE",
                0x2000: "RTM_F_FIB_MATCH",
                0x4000: "RTM_F_OFFLOAD",
                0x8000: "RTM_F_TRAP",
                0x20000000: "RTM_F_OFFLOAD_FAILED",
            },
        ),
        # Pay
        PacketListField("data", [], rtmsg_rtattr),
    ]


bind_layers(rtmsghdr, rtmsg, nlmsg_type=24)
bind_layers(rtmsghdr, rtmsg, nlmsg_type=25)
bind_layers(rtmsghdr, rtmsg, nlmsg_type=26)


class rtmsghdrs(Packet):
    fields_desc = [
        PacketListField(
            "msgs",
            [],
            rtmsghdr,
            # 65535 / len(rtmsghdr)
            max_count=4096,
        ),
    ]


# Utils


SOL_NETLINK = 270
NETLINK_EXT_ACK = 11
NETLINK_GET_STRICT_CHK = 12


def _sr1_rtrequest(pkt: Packet) -> List[Packet]:
    """
    Send / Receive a rtnetlink request
    """
    # Create socket
    sock = socket.socket(
        socket.AF_NETLINK,
        socket.SOCK_RAW | socket.SOCK_CLOEXEC,
        socket.NETLINK_ROUTE,
    )
    # Configure socket
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 32768)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1048576)
    try:
        sock.setsockopt(SOL_NETLINK, NETLINK_EXT_ACK, 1)
    except OSError:
        # Linux 4.12+ only
        pass
    sock.bind((0, 0))  # bind to kernel
    try:
        sock.setsockopt(SOL_NETLINK, NETLINK_GET_STRICT_CHK, 1)
    except OSError:
        # Linux 4.20+ only
        pass
    # Request routes
    sock.send(bytes(rtmsghdrs(msgs=[pkt])))
    results: List[Packet] = []
    try:
        while True:
            msgs = rtmsghdrs(sock.recv(65535))
            if not msgs:
                log_loading.warning("Failed to read the routes using RTNETLINK !")
                return []
            for msg in msgs.msgs:
                # Keep going until we find the end of the MULTI format
                if not msg.nlmsg_flags.NLM_F_MULTI or msg.nlmsg_type == 3:
                    if msg.nlmsg_type == 3 and nlmsgerr in msg and msg.status != 0:
                        # NLMSG_DONE with errors
                        if msg.data and msg.data[0].rta_type == 1:
                            log_loading.debug(
                                "Scapy RTNETLINK error on %s: '%s'. Please report !",
                                pkt.sprintf("%nlmsg_type%"),
                                msg.data[0].rta_data.decode(),
                            )
                            return []
                    return results
                results.append(msg)
    finally:
        sock.close()


def _get_ips(af_family=socket.AF_UNSPEC):
    # type: (socket.AddressFamily) -> Dict[int, List[Dict[str, Any]]]
    """
    Return a mapping of all interfaces IP using a NETLINK socket.
    """
    results = _sr1_rtrequest(
        rtmsghdr(
            nlmsg_type="RTM_GETADDR",
            nlmsg_flags="NLM_F_REQUEST+NLM_F_ROOT+NLM_F_MATCH",
            nlmsg_seq=int(time.time()),
        )
        / ifaddrmsg(
            ifa_family=af_family,
            data=[],
        )
    )
    ips: Dict[int, List[Dict[str, Any]]] = {}
    for msg in results:
        ifindex = msg.ifa_index
        address = None
        family = msg.ifa_family
        local = None
        for attr in msg.data:
            if attr.rta_type == 0x01:  # IFA_ADDRESS
                address = attr.rta_data
            elif attr.rta_type == 0x02:  # IFA_LOCAL
                local = attr.rta_data
        # include/uapi/linux/if_addr.h: for point-to-point links, IFA_LOCAL is the local
        # interface address and IFA_ADDRESS is the peer address
        local_address = local if local is not None else address
        if local_address is not None:
            data = {
                "af_family": family,
                "index": ifindex,
                "address": local_address,
            }
            if family == 10:  # ipv6
                data["scope"] = scapy.utils6.in6_getscope(local_address)
            ips.setdefault(ifindex, list()).append(data)
    return ips


def _get_if_list():
    # type: () -> Dict[int, Dict[str, Any]]
    """
    Read the interfaces list using a NETLINK socket.
    """
    results = _sr1_rtrequest(
        rtmsghdr(
            nlmsg_type="RTM_GETLINK",
            nlmsg_flags="NLM_F_REQUEST+NLM_F_ROOT+NLM_F_MATCH",
            nlmsg_seq=int(time.time()),
        )
        / ifinfomsg(
            data=[],
        )
    )
    lifips = _get_ips()
    interfaces = {}
    for msg in results:
        ifindex = msg.ifi_index
        ifname = None
        mac = "00:00:00:00:00:00"
        itype = msg.ifi_type
        ifflags = msg.ifi_flags
        ips = []
        for attr in msg.data:
            if attr.rta_type == 0x01:  # IFLA_ADDRESS
                mac = attr.rta_data
            elif attr.rta_type == 0x03:  # IFLA_NAME
                ifname = attr.rta_data[:-1].decode()
        if ifname is not None:
            if ifindex in lifips:
                ips = lifips[ifindex]
            interfaces[ifindex] = {
                "name": ifname,
                "index": ifindex,
                "flags": ifflags,
                "mac": mac,
                "type": itype,
                "ips": ips,
            }
    return interfaces


def in6_getifaddr():
    # type: () -> List[Tuple[str, int, str]]
    """
    Returns a list of 3-tuples of the form (addr, scope, iface) where
    'addr' is the address of scope 'scope' associated to the interface
    'iface'.

    This is the list of all addresses of all interfaces available on
    the system.
    """
    ips = _get_ips(af_family=socket.AF_INET6)
    ifaces = _get_if_list()
    result = []
    for intip in ips.values():
        for ip in intip:
            if ip["index"] in ifaces:
                result.append((ip["address"], ip["scope"], ifaces[ip["index"]]["name"]))
    return result


def _read_routes(af_family):
    # type: (socket.AddressFamily) -> List[Packet]
    """
    Read routes using a NETLINK socket.
    """
    results = []
    for rttable in ["RT_TABLE_LOCAL", "RT_TABLE_MAIN"]:
        results.extend(
            _sr1_rtrequest(
                rtmsghdr(
                    nlmsg_type="RTM_GETROUTE",
                    nlmsg_flags="NLM_F_REQUEST+NLM_F_ROOT+NLM_F_MATCH",
                    nlmsg_seq=int(time.time()),
                )
                / rtmsg(
                    rtm_family=af_family,
                    data=[
                        rtmsg_rtattr(rta_type="RTA_TABLE", rta_data=rttable),
                    ],
                )
            )
        )
    return [msg for msg in results if msg.nlmsg_type == 24]  # RTM_NEWROUTE


def read_routes():
    # type: () -> List[Tuple[int, int, str, str, str, int]]
    """
    Read IPv4 routes for current process
    """
    routes = []
    ifaces = _get_if_list()
    results = _read_routes(socket.AF_INET)
    for msg in results:
        # Omit stupid answers (some OS conf appears to lead to this)
        if msg.rtm_family != socket.AF_INET:
            continue
        # Process the RTM_NEWROUTE
        net = 0
        mask = itom(msg.rtm_dst_len)
        gw = "0.0.0.0"
        iface = ""
        addr = "0.0.0.0"
        metric = 0
        for attr in msg.data:
            if attr.rta_type == 0x01:  # RTA_DST
                net = atol(attr.rta_data)
            elif attr.rta_type == 0x04:  # RTS_OIF
                index = attr.rta_data
                if index in ifaces:
                    iface = ifaces[index]["name"]
                else:
                    iface = str(index)
            elif attr.rta_type == 0x05:  # RTA_GATEWAY
                gw = attr.rta_data
            elif attr.rta_type == 0x06:  # RTA_PRIORITY
                metric = attr.rta_data
            elif attr.rta_type == 0x07:  # RTA_PREFSRC
                addr = attr.rta_data
        routes.append((net, mask, gw, iface, addr, metric))
    # Add multicast routes, as those are missing by default
    for _iface in ifaces.values():
        if _iface['flags'].MULTICAST:
            try:
                addr = next(
                    x["address"]
                    for x in _iface["ips"]
                    if x["af_family"] == socket.AF_INET
                )
            except StopIteration:
                continue
            routes.append((
                0xe0000000, 0xf0000000, "0.0.0.0", _iface["name"], addr, 250
            ))
    return routes


def read_routes6():
    # type: () -> List[Tuple[str, int, str, str, List[str], int]]
    """
    Read IPv6 routes for current process
    """
    routes = []
    ifaces = _get_if_list()
    results = _read_routes(socket.AF_INET6)
    lifaddr = _get_ips(af_family=socket.AF_INET6)
    for msg in results:
        # Omit stupid answers (some OS conf appears to lead to this)
        if msg.rtm_family != socket.AF_INET6:
            continue
        # Process the RTM_NEWROUTE
        prefix = "::"
        plen = msg.rtm_dst_len
        nh = "::"
        index = 0
        iface = ""
        metric = 0
        for attr in msg.data:
            if attr.rta_type == 0x01:  # RTA_DST
                prefix = attr.rta_data
            elif attr.rta_type == 0x04:  # RTS_OIF
                index = attr.rta_data
                if index in ifaces:
                    iface = ifaces[index]["name"]
                else:
                    iface = str(index)
            elif attr.rta_type == 0x05:  # RTA_GATEWAY
                nh = attr.rta_data
            elif attr.rta_type == 0x06:  # RTA_PRIORITY
                metric = attr.rta_data
        devaddrs = ((x["address"], x["scope"], iface) for x in lifaddr.get(index, []))
        cset = scapy.utils6.construct_source_candidate_set(prefix, plen, devaddrs)
        if cset:
            routes.append((prefix, plen, nh, iface, cset, metric))
    # Add multicast routes, as those are missing by default
    for _iface in ifaces.values():
        if _iface['flags'].MULTICAST:
            addrs = [
                x["address"]
                for x in _iface["ips"]
                if x["af_family"] == socket.AF_INET6
            ]
            if not addrs:
                continue
            routes.append((
                "ff00::", 8, "::", _iface["name"], addrs, 250
            ))
    return routes



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = UDS Ecu logging additions
# scapy.contrib.status = library

from scapy.contrib.automotive.uds import UDS_DSCPR, UDS_ERPR, UDS_SAPR, \
    UDS_CCPR, UDS_TPPR, UDS_DSC, UDS_ER, UDS_RDPR, UDS_TDPR, UDS_RD, UDS_TD, \
    UDS_CC, UDS_NR, UDS_SA, UDS_RDBIPR, UDS_LC, UDS_RC, UDS_TP, UDS_RU, \
    UDS_IOCBIPR, UDS_WDBIPR, UDS_CDTCIPR, UDS_CDTCI, UDS_RDTCIPR, \
    UDS_RDTCI, UDS_RMBAPR, UDS_WMBAPR, UDS_WMBA, UDS_LCPR, UDS_RCPR, UDS_RFT, \
    UDS_RTE, UDS_RTEPR, UDS_RFTPR, UDS_IOCBI, UDS_RDBI, UDS_RMBA, UDS_WDBI, \
    UDS_CDTCS, UDS_CDTCSPR, UDS_SDT, UDS_SDTPR, UDS_RUPR
from scapy.packet import Packet
from scapy.contrib.automotive.ecu import Ecu

from typing import (
    Any,
    Tuple,
)


@Ecu.extend_pkt_with_logging(UDS_DSC)
def UDS_DSC_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_DSC.diagnosticSessionType%")


@Ecu.extend_pkt_with_logging(UDS_DSCPR)
def UDS_DSCPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_DSCPR.diagnosticSessionType%")


@Ecu.extend_pkt_with_logging(UDS_ER)
def UDS_ER_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_ER.resetType%")


@Ecu.extend_pkt_with_logging(UDS_ERPR)
def UDS_ERPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_ER.resetType%")


@Ecu.extend_pkt_with_logging(UDS_SA)
def UDS_SA_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    if self.securityAccessType % 2 == 1:
        return self.sprintf("%UDS.service%"),\
            (self.securityAccessType, None)
    else:
        return self.sprintf("%UDS.service%"),\
            (self.securityAccessType, self.securityKey)


@Ecu.extend_pkt_with_logging(UDS_SAPR)
def UDS_SAPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    if self.securityAccessType % 2 == 0:
        return self.sprintf("%UDS.service%"),\
            (self.securityAccessType, None)
    else:
        return self.sprintf("%UDS.service%"),\
            (self.securityAccessType, self.securitySeed)


@Ecu.extend_pkt_with_logging(UDS_CC)
def UDS_CC_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_CC.controlType%")


@Ecu.extend_pkt_with_logging(UDS_CCPR)
def UDS_CCPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_CCPR.controlType%")


@Ecu.extend_pkt_with_logging(UDS_TP)
def UDS_TP_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.subFunction


@Ecu.extend_pkt_with_logging(UDS_TPPR)
def UDS_TPPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.zeroSubFunction


@Ecu.extend_pkt_with_logging(UDS_SDT)
def UDS_SDT_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.securityDataRequestRecord


@Ecu.extend_pkt_with_logging(UDS_SDTPR)
def UDS_SDTPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.securityDataResponseRecord


@Ecu.extend_pkt_with_logging(UDS_CDTCS)
def UDS_CDTCS_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_CDTCS.DTCSettingType%")


@Ecu.extend_pkt_with_logging(UDS_CDTCSPR)
def UDS_CDTCSPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_CDTCSPR.DTCSettingType%")


@Ecu.extend_pkt_with_logging(UDS_LC)
def UDS_LC_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS.linkControlType%")


@Ecu.extend_pkt_with_logging(UDS_LCPR)
def UDS_LCPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS.linkControlType%")


@Ecu.extend_pkt_with_logging(UDS_RDBI)
def UDS_RDBI_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_RDBI.identifiers%")


@Ecu.extend_pkt_with_logging(UDS_RDBIPR)
def UDS_RDBIPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_RDBIPR.dataIdentifier%")


@Ecu.extend_pkt_with_logging(UDS_RMBA)
def UDS_RMBA_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        (getattr(self, "memoryAddress%d" % self.memoryAddressLen),
         getattr(self, "memorySize%d" % self.memorySizeLen))


@Ecu.extend_pkt_with_logging(UDS_RMBAPR)
def UDS_RMBAPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.dataRecord


@Ecu.extend_pkt_with_logging(UDS_WDBI)
def UDS_WDBI_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_WDBI.dataIdentifier%")


@Ecu.extend_pkt_with_logging(UDS_WDBIPR)
def UDS_WDBIPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        self.sprintf("%UDS_WDBIPR.dataIdentifier%")


@Ecu.extend_pkt_with_logging(UDS_WMBA)
def UDS_WMBA_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    addr = getattr(self, "memoryAddress%d" % self.memoryAddressLen)
    size = getattr(self, "memorySize%d" % self.memorySizeLen)
    return self.sprintf("%UDS.service%"), (addr, size, self.dataRecord)


@Ecu.extend_pkt_with_logging(UDS_WMBAPR)
def UDS_WMBAPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    addr = getattr(self, "memoryAddress%d" % self.memoryAddressLen)
    size = getattr(self, "memorySize%d" % self.memorySizeLen)
    return self.sprintf("%UDS.service%"), (addr, size)


@Ecu.extend_pkt_with_logging(UDS_CDTCI)
def UDS_CDTCI_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        (self.groupOfDTCHighByte, self.groupOfDTCMiddleByte,
         self.groupOfDTCLowByte)


@Ecu.extend_pkt_with_logging(UDS_CDTCIPR)
def UDS_CDTCIPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), None


@Ecu.extend_pkt_with_logging(UDS_RDTCI)
def UDS_RDTCI_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), repr(self)


@Ecu.extend_pkt_with_logging(UDS_RDTCIPR)
def UDS_RDTCIPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), repr(self)


@Ecu.extend_pkt_with_logging(UDS_RC)
def UDS_RC_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"),\
        (self.routineControlType,
         self.routineIdentifier)


@Ecu.extend_pkt_with_logging(UDS_RCPR)
def UDS_RCPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"),\
        (self.routineControlType,
         self.routineIdentifier)


@Ecu.extend_pkt_with_logging(UDS_RD)
def UDS_RD_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    addr = getattr(self, "memoryAddress%d" % self.memoryAddressLen)
    size = getattr(self, "memorySize%d" % self.memorySizeLen)
    return self.sprintf("%UDS.service%"), (addr, size)


@Ecu.extend_pkt_with_logging(UDS_RDPR)
def UDS_RDPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.memorySizeLen


@Ecu.extend_pkt_with_logging(UDS_RU)
def UDS_RU_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    addr = getattr(self, "memoryAddress%d" % self.memoryAddressLen)
    size = getattr(self, "memorySize%d" % self.memorySizeLen)
    return self.sprintf("%UDS.service%"), (addr, size)


@Ecu.extend_pkt_with_logging(UDS_RUPR)
def UDS_RUPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.memorySizeLen


@Ecu.extend_pkt_with_logging(UDS_TD)
def UDS_TD_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"),\
        (self.blockSequenceCounter, self.transferRequestParameterRecord)


@Ecu.extend_pkt_with_logging(UDS_TDPR)
def UDS_TDPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.blockSequenceCounter


@Ecu.extend_pkt_with_logging(UDS_RTE)
def UDS_RTE_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"),\
        self.transferRequestParameterRecord


@Ecu.extend_pkt_with_logging(UDS_RTEPR)
def UDS_RTEPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"),\
        self.transferResponseParameterRecord


@Ecu.extend_pkt_with_logging(UDS_RFT)
def UDS_RFT_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"),\
        self.modeOfOperation


@Ecu.extend_pkt_with_logging(UDS_RFTPR)
def UDS_RFTPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"),\
        self.modeOfOperation


@Ecu.extend_pkt_with_logging(UDS_IOCBI)
def UDS_IOCBI_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.dataIdentifier


@Ecu.extend_pkt_with_logging(UDS_IOCBIPR)
def UDS_IOCBIPR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), self.dataIdentifier


@Ecu.extend_pkt_with_logging(UDS_NR)
def UDS_NR_get_log(self):
    # type: (Packet) -> Tuple[str, Any]
    return self.sprintf("%UDS.service%"), \
        (self.sprintf("%UDS_NR.requestServiceId%"),
         self.sprintf("%UDS_NR.negativeResponseCode%"))



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Pavel Oborin <oborin.p@gmail.com>

# RFC 3550
# scapy.contrib.description = Real-Time Transport Control Protocol
# scapy.contrib.status = loads

"""
RTCP (rfc 3550)

Use bind_layers(UDP, RTCP, dport=...) to start using it
"""

import struct

from scapy.packet import Packet
from scapy.fields import (
    BitField,
    BitFieldLenField,
    ByteEnumField,
    ByteField,
    ConditionalField,
    FieldLenField,
    IntField,
    LenField,
    LongField,
    PacketField,
    PacketListField,
    StrLenField,
    X3BytesField,
)


_rtcp_packet_types = {
    200: 'Sender report',
    201: 'Receiver report',
    202: 'Source description',
    203: 'BYE',
    204: 'APP'
}


class SenderInfo(Packet):
    name = "Sender info"
    fields_desc = [
        LongField('ntp_timestamp', None),
        IntField('rtp_timestamp', None),
        IntField('sender_packet_count', None),
        IntField('sender_octet_count', None)
    ]

    def extract_padding(self, p):
        return "", p


class ReceptionReport(Packet):
    name = "Reception report"
    fields_desc = [
        IntField('sourcesync', None),
        ByteField('fraction_lost', None),
        X3BytesField('cumulative_lost', None),
        IntField('highest_seqnum_recv', None),
        IntField('interarrival_jitter', None),
        IntField('last_SR_timestamp', None),
        IntField('delay_since_last_SR', None)
    ]

    def extract_padding(self, p):
        return "", p


_sdes_chunk_types = {
    0: "END",
    1: "CNAME",
    2: "NAME",
    3: "EMAIL",
    4: "PHONE",
    5: "LOC",
    6: "TOOL",
    7: "NOTE",
    8: "PRIV"
}


class SDESItem(Packet):
    name = "SDES item"
    fields_desc = [
        ByteEnumField('chunk_type', None, _sdes_chunk_types),
        FieldLenField('length', None, fmt='!b', length_of='value'),
        StrLenField('value', None, length_from=lambda pkt: pkt.length)
    ]

    def extract_padding(self, p):
        return "", p


class SDESChunk(Packet):
    name = "SDES chunk"
    fields_desc = [
        IntField('sourcesync', None),
        PacketListField(
            'items', None,
            next_cls_cb=(
                lambda x, y, p, z: None if (p and p.chunk_type == 0) else SDESItem
            )
        )
    ]


class RTCP(Packet):
    name = "RTCP"

    fields_desc = [
        # HEADER
        BitField('version', 2, 2),
        BitField('padding', 0, 1),
        BitFieldLenField('count', 0, 5, count_of='report_blocks'),
        ByteEnumField('packet_type', 0, _rtcp_packet_types),
        LenField('length', None, fmt='!h'),
        # SR/RR
        ConditionalField(
            IntField('sourcesync', 0),
            lambda pkt: pkt.packet_type in (200, 201)
        ),
        ConditionalField(
            PacketField('sender_info', SenderInfo(), SenderInfo),
            lambda pkt: pkt.packet_type == 200
        ),
        ConditionalField(
            PacketListField('report_blocks', None, pkt_cls=ReceptionReport,
                            count_from=lambda pkt: pkt.count),
            lambda pkt: pkt.packet_type in (200, 201)
        ),
        # SDES
        ConditionalField(
            PacketListField('sdes_chunks', None, pkt_cls=SDESChunk,
                            count_from=lambda pkt: pkt.count),
            lambda pkt: pkt.packet_type == 202
        ),
    ]

    def post_build(self, pkt, pay):
        pkt += pay
        if self.length is None:
            pkt = pkt[:2] + struct.pack("!h", len(pkt) // 4 - 1) + pkt[4:]
        return pkt



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Pavel Oborin <oborin.p@gmail.com>

# RFC 8489
# scapy.contrib.description = Session Traversal Utilities for NAT (STUN)
# scapy.contrib.status = loads

"""
    STUN (RFC 8489)

    TLV code derived from the DTP implementation:
      Thanks to Nicolas Bareil,
                Arnaud Ebalard,
                Jochen Bartl.
"""
import struct
import itertools

from scapy.layers.inet import UDP, TCP
from scapy.config import conf
from scapy.packet import Packet, bind_bottom_up, bind_top_down
from scapy.utils import inet_ntoa, inet_aton
from scapy.fields import (
    BitField,
    BitEnumField,
    LenField,
    IntField,
    PadField,
    StrLenField,
    PacketListField,
    XShortField,
    FieldLenField,
    ShortField,
    ByteEnumField,
    ByteField,
    XNBytesField,
    XLongField,
    XIntField,
    XBitField,
    IPField,
    IP6Field,
    MultipleTypeField,
)

MAGIC_COOKIE = 0x2112A442

_stun_class = {
    "request": 0b00,
    "indication": 0b01,
    "success response": 0b10,
    "error response": 0b11
}

_stun_method = {
    "Binding": 0b000000000001
}

# fmt: off
_stun_message_type = {
    "{} {}".format(method, class_):
        (method_code & 0b000000001111)      |    # noqa: E221,W504
        (class_code  & 0b01)           << 4 |    # noqa: E221,W504
        (method_code & 0b000001110000) << 5 |    # noqa: E221,W504
        (class_code  & 0b10)           << 7 |    # noqa: E221,W504
        (method_code & 0b111110000000) << 9
    for (method, method_code), (class_, class_code) in
        itertools.product(_stun_method.items(), _stun_class.items())    # noqa: E131
}
# fmt: on


class STUNGenericTlv(Packet):
    name = "STUN Generic TLV"

    fields_desc = [
        XShortField("type", 0x0000),
        FieldLenField("length", None, length_of="value"),
        PadField(StrLenField("value", "", length_from=lambda pkt:pkt.length), align=4)
    ]

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kwargs):
        if _pkt and len(_pkt) >= 2:
            t = struct.unpack("!H", _pkt[:2])[0]
            return _stun_tlv_class.get(t, cls)
        return cls

    def guess_payload_class(self, payload):
        return conf.padding_layer


class STUNUsername(STUNGenericTlv):
    name = "STUN Username"

    fields_desc = [
        XShortField("type", 0x0006),
        FieldLenField("length", None, length_of="username"),
        PadField(
            StrLenField("username", '', length_from=lambda pkt: pkt.length),
            align=4, padwith=b"\x20"
        )
    ]


class STUNMessageIntegrity(STUNGenericTlv):
    name = "STUN Message Integrity"

    fields_desc = [
        XShortField("type", 0x0008),
        ShortField("length", 20),
        XNBytesField("hmac_sha1", 0, 20)
    ]

    def post_build(self, pkt, pay):
        pkt += pay
        return pkt


class STUNPriority(STUNGenericTlv):
    name = "STUN Priority"

    fields_desc = [
        XShortField("type", 0x0024),
        ShortField("length", 4),
        IntField("priority", 0)
    ]


_xor_mapped_address_family = {
    "IPv4": 0x01,
    "IPv6": 0x02
}


class XorPort(ShortField):

    def m2i(self, pkt, x):
        return x ^ (MAGIC_COOKIE >> 16)

    def i2m(self, pkt, x):
        return x ^ (MAGIC_COOKIE >> 16)


class XorIp(IPField):

    def m2i(self, pkt, x):
        return inet_ntoa(struct.pack(">i", (struct.unpack(">i", x)[0] ^ MAGIC_COOKIE)))

    def i2m(self, pkt, x):
        if x is None:
            return b"\x00\x00\x00\x00"
        return struct.pack(">i", struct.unpack(">i", inet_aton(x))[0] ^ MAGIC_COOKIE)


class XorIp6(IP6Field):

    def m2i(self, pkt, x):
        addr = self._xor_address(pkt, x)
        return super().m2i(pkt, addr)

    def i2m(self, pkt, x):
        addr = super().i2m(pkt, x)
        return self._xor_address(pkt, addr)

    def _xor_address(self, pkt, addr):
        xor_words = [pkt.parent.magic_cookie]
        xor_words += struct.unpack(
            ">III", pkt.parent.transaction_id.to_bytes(12, "big")
        )
        addr_words = struct.unpack(">IIII", addr)
        xor_addr = [a ^ b for a, b in zip(addr_words, xor_words)]
        return struct.pack(">IIII", *xor_addr)


class STUNXorMappedAddress(STUNGenericTlv):
    name = "STUN XOR Mapped Address"

    fields_desc = [
        XShortField("type", 0x0020),
        FieldLenField("length", None, length_of="xip", adjust=lambda pkt, x: x + 4),
        ByteField("RESERVED", 0),
        ByteEnumField("address_family", 1, _xor_mapped_address_family),
        XorPort("xport", 0),
        MultipleTypeField(
            [
                (XorIp("xip", "127.0.0.1"), lambda pkt: pkt.address_family == 1),
                (XorIp6("xip", "::1"), lambda pkt: pkt.address_family == 2),
            ],
            XorIp("xip", "127.0.0.1"),
        ),
    ]


class STUNMappedAddress(STUNGenericTlv):
    name = "STUN Mapped Address"

    fields_desc = [
        XShortField("type", 0x0001),
        FieldLenField("length", None, length_of="ip", adjust=lambda pkt, x: x + 4),
        ByteField("RESERVED", 0),
        ByteEnumField("address_family", 1, _xor_mapped_address_family),
        ShortField("port", 0),
        MultipleTypeField(
            [
                (IPField("ip", "127.0.0.1"), lambda pkt: pkt.address_family == 1),
                (IP6Field("ip", "::1"), lambda pkt: pkt.address_family == 2),
            ],
            IPField("ip", "127.0.0.1"),
        ),
    ]


class STUNUseCandidate(STUNGenericTlv):
    name = "STUN Use Candidate"

    fields_desc = [
        XShortField("type", 0x0025),
        FieldLenField("length", 0, length_of="value"),
        PadField(StrLenField("value", "", length_from=lambda pkt: pkt.length), align=4)
    ]


class STUNFingerprint(STUNGenericTlv):
    name = "STUN Fingerprint"

    fields_desc = [
        XShortField("type", 0x8028),
        ShortField("length", 4),
        XIntField("crc_32", None)
    ]


class STUNIceControlling(STUNGenericTlv):
    name = "STUN ICE-controlling"

    fields_desc = [
        XShortField("type", 0x802a),
        ShortField("length", 8),
        XLongField("tie_breaker", None)
    ]


class STUNGoogNetworkInfo(STUNGenericTlv):
    name = "STUN Google Network Info"

    fields_desc = [
        XShortField("type", 0xc057),
        ShortField("length", 4),
        ShortField("network_id", 0),
        ShortField("network_cost", 999)
    ]


_stun_tlv_class = {
    0x0001: STUNMappedAddress,
    0x0006: STUNUsername,
    0x0008: STUNMessageIntegrity,
    0x0020: STUNXorMappedAddress,
    0x0025: STUNUseCandidate,
    0x0024: STUNPriority,
    0x8028: STUNFingerprint,
    0x802a: STUNIceControlling,
    0xc057: STUNGoogNetworkInfo
}

_stun_tlv_attribute_types = {
    "MAPPED-ADDRESS": 0x0001,
    "USERNAME": 0x0006,
    "MESSAGE-INTEGRITY": 0x0008,
    "ERROR-CODE": 0x0009,
    "UNKNOWN-ATTRIBUTES": 0x000A,
    "REALM": 0x0014,
    "NONCE": 0x0015,
    "XOR-MAPPED-ADDRESS": 0x0020,
    "PRIORITY": 0x0024,
    "USE-CANDIDATE": 0x0025,
    "SOFTWARE": 0x8022,
    "ALTERNATE-SERVER": 0x8023,
    "FINGERPRINT": 0x8028,
    "ICE-CONTROLLED": 0x8029,
    "ICE-CONTROLLING": 0x802a,
    "GOOG-NETWORK-INFO": 0xc057
}


class STUN(Packet):
    description = ""

    fields_desc = [
        BitField('RESERVED', 0b00, size=2),   # <- always zeroes
        BitEnumField('stun_message_type', None, 14, _stun_message_type),
        LenField('length', None, fmt='!h'),
        XIntField('magic_cookie', MAGIC_COOKIE),
        XBitField('transaction_id', None, 96),
        PacketListField("attributes", [], STUNGenericTlv)
    ]

    def post_build(self, pkt, pay):
        pkt += pay
        if self.length is None:
            pkt = pkt[:2] + struct.pack("!h", len(pkt) - 20) + pkt[4:]
        for attr in self.attributes:
            if isinstance(attr, STUNMessageIntegrity):
                pass    # TODO Fill hmac-sha1 in MESSAGE-INTEGRITY attribute
        return pkt


bind_bottom_up(UDP, STUN, sport=3478)
bind_bottom_up(UDP, STUN, dport=3478)
bind_top_down(UDP, STUN, sport=3478, dport=3478)

bind_bottom_up(TCP, STUN, sport=3478)
bind_bottom_up(TCP, STUN, dport=3478)
bind_top_down(TCP, STUN, sport=3478, dport=3478)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright 2012 (C) The MITRE Corporation

"""
.. centered::
    NOTICE
    This software/technical data was produced for the U.S. Government
    under Prime Contract No. NASA-03001 and JPL Contract No. 1295026
    and is subject to FAR 52.227-14 (6/87) Rights in Data General,
    and Article GP-51, Rights in Data  General, respectively.
    This software is publicly released under MITRE case #12-3054
"""

# scapy.contrib.description = Licklider Transmission Protocol (LTP)
# scapy.contrib.status = loads

from scapy.packet import Packet, bind_layers, bind_top_down
from scapy.fields import BitEnumField, BitField, BitFieldLenField, \
    ByteEnumField, ConditionalField, PacketListField, StrLenField
from scapy.layers.inet import UDP
from scapy.config import conf
from scapy.contrib.sdnv import SDNV2, SDNV2FieldLenField

# LTP https://tools.ietf.org/html/rfc5326

_ltp_flag_vals = {
    0: '0x0 Red data, NOT (Checkpoint, EORP or EOB)',
    1: '0x1 Red data, Checkpoint, NOT (EORP or EOB)',
    2: '0x2 Red data, Checkpoint, EORP, NOT EOB',
    3: '0x3 Red data, Checkpoint, EORP, EOB',
    4: '0x4 Green data, NOT EOB',
    5: '0x5 Green data, undefined',
    6: '0x6 Green data, undefined',
    7: '0x7 Green data, EOB',
    8: '0x8 Report segment',
    9: '0x9 Report-acknowledgment segmen',
    10: '0xA Control segment, undefined',
    11: '0xB Control segment, undefined',
    12: '0xC Cancel segment from block sender',
    13: '0xD Cancel-acknowledgment segment to block sender',
    14: '0xE Cancel segment from block receiver',
    15: '0xF Cancel-acknowledgment segment to block receiver'}

_ltp_cancel_reasons = {
    0: 'USR_CNCLD  - Client service canceled session.',
    1: 'UNREACH    - Unreachable client service.',
    2: 'RLEXC      - Retransmission limit exceeded.',
    3: 'MISCOLORED - Received miscolored segment.',
    4: 'SYS_CNCLD  - System error condition.',
    5: 'RXMTCYCEXC - Exceeded the retransmission cycles limit.',
    6: 'RESERVED'}   # Reserved 0x06-0xFF

# LTP Extensions https://tools.ietf.org/html/rfc5327

_ltp_extension_tag = {
    0: 'LTP authentication extension',
    1: 'LTP cookie extension'
}

_ltp_data_segment = [0, 1, 2, 3, 4, 5, 6, 7]
_ltp_checkpoint_segment = [1, 2, 3]

_ltp_payload_conditions = {}


def ltp_bind_payload(cls, lambd):
    """Bind payload class to the LTP packets.

    :param cls: the class to bind
    :param lambd: lambda that will be called to check
        whether or not the cls should be used
        ex: lambda pkt: ...
    """
    _ltp_payload_conditions[cls] = lambd


class LTPex(Packet):
    name = "LTP Extension"
    fields_desc = [
        ByteEnumField("ExTag", 0, _ltp_extension_tag),
        SDNV2FieldLenField("ExLength", None, length_of="ExData"),
        # SDNV2FieldLenField
        StrLenField("ExData", "", length_from=lambda x: x.ExLength)
    ]

    def default_payload_class(self, pay):
        return conf.padding_layer


class LTPReceptionClaim(Packet):
    name = "LTP Reception Claim"
    fields_desc = [SDNV2("ReceptionClaimOffset", 0),
                   SDNV2("ReceptionClaimLength", 0)]

    def default_payload_class(self, pay):
        return conf.padding_layer


def _ltp_guess_payload(pkt, *args):
    for k, v in _ltp_payload_conditions.items():
        if v(pkt):
            return k
    return conf.raw_layer


class LTP(Packet):
    name = "LTP"
    fields_desc = [
        BitField('version', 0, 4),
        BitEnumField('flags', 0, 4, _ltp_flag_vals),
        SDNV2("SessionOriginator", 0),
        SDNV2("SessionNumber", 0),
        BitFieldLenField("HeaderExtensionCount", None, 4, count_of="HeaderExtensions"),  # noqa: E501
        BitFieldLenField("TrailerExtensionCount", None, 4, count_of="TrailerExtensions"),  # noqa: E501
        PacketListField("HeaderExtensions", [], LTPex, count_from=lambda x: x.HeaderExtensionCount),  # noqa: E501
        #
        # LTP segments containing data have a DATA header
        #
        ConditionalField(SDNV2("DATA_ClientServiceID", 0),
                         lambda x: x.flags in _ltp_data_segment),
        ConditionalField(SDNV2("DATA_PayloadOffset", 0),
                         lambda x: x.flags in _ltp_data_segment),
        ConditionalField(SDNV2FieldLenField("DATA_PayloadLength", None, length_of="LTP_Payload"),  # noqa: E501
                         lambda x: x.flags in _ltp_data_segment),
        #
        # LTP segments that are checkpoints will have a checkpoint serial number and report serial number.  # noqa: E501
        #
        ConditionalField(SDNV2("CheckpointSerialNo", 0),
                         lambda x: x.flags in _ltp_checkpoint_segment),
        #
        # For segments that are checkpoints or reception reports.
        #
        ConditionalField(SDNV2("ReportSerialNo", 0),
                         lambda x: x.flags in _ltp_checkpoint_segment \
                         or x.flags == 8),
        #
        # Then comes the actual payload for data carrying segments.
        #
        ConditionalField(PacketListField("LTP_Payload", None, next_cls_cb=_ltp_guess_payload,  # noqa: E501
                                         length_from=lambda x: x.DATA_PayloadLength),  # noqa: E501
                         lambda x: x.flags in _ltp_data_segment),
        #
        # Report ACKS acknowledge a particular report serial number.
        #
        ConditionalField(SDNV2("RA_ReportSerialNo", 0),
                         lambda x: x.flags == 9),
        #
        # Reception reports have the following fields,
        # excluding ReportSerialNo defined above.
        #
        ConditionalField(SDNV2("ReportCheckpointSerialNo", 0),
                         lambda x: x.flags == 8),
        ConditionalField(SDNV2("ReportUpperBound", 0),
                         lambda x: x.flags == 8),
        ConditionalField(SDNV2("ReportLowerBound", 0),
                         lambda x: x.flags == 8),
        ConditionalField(SDNV2FieldLenField("ReportReceptionClaimCount", None, count_of="ReportReceptionClaims"),  # noqa: E501
                         lambda x: x.flags == 8),
        ConditionalField(PacketListField("ReportReceptionClaims", [], LTPReceptionClaim,  # noqa: E501
                                         count_from=lambda x: x.ReportReceptionClaimCount),  # noqa: E501
                         lambda x: x.flags == 8 and (not x.ReportReceptionClaimCount or x.ReportReceptionClaimCount > 0)),  # noqa: E501
        #
        # Cancellation Requests
        #
        ConditionalField(ByteEnumField("CancelFromSenderReason",
                                       15, _ltp_cancel_reasons),
                         lambda x: x.flags == 12),
        ConditionalField(ByteEnumField("CancelFromReceiverReason",
                                       15, _ltp_cancel_reasons),
                         lambda x: x.flags == 14),
        #
        # Finally, trailing extensions
        #
        PacketListField("TrailerExtensions", [], LTPex, count_from=lambda x: x.TrailerExtensionCount)  # noqa: E501
    ]

    def mysummary(self):
        return self.sprintf("LTP %SessionNumber%"), [UDP]


bind_top_down(UDP, LTP, sport=1113)
bind_top_down(UDP, LTP, dport=1113)
bind_top_down(UDP, LTP, sport=2113)
bind_top_down(UDP, LTP, dport=2113)
bind_layers(UDP, LTP, sport=1113, dport=1113)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2018 Francois Contat <francois.contat@ssi.gouv.fr>

"""
RTR

Based on RTR RFC 6810 https://tools.ietf.org/html/rfc6810 for version 0
Based on RTR RFC 8210 https://tools.ietf.org/html/rfc8210 for version 1
"""

# scapy.contrib.description = The RPKI to Router Protocol
# scapy.contrib.status = loads

# Start dev

import struct

from scapy.packet import Packet, bind_layers, Raw
from scapy.fields import ByteEnumField, ByteField, IntField, ShortField
from scapy.fields import IPField, IP6Field, StrLenField
from scapy.fields import FieldLenField
from scapy.fields import StrFixedLenField, ShortEnumField
from scapy.layers.inet import TCP
from scapy.compat import orb

STATIC_SERIAL_NOTIFY_LENGTH = 12
STATIC_SERIAL_QUERY_LENGTH = 12
STATIC_RESET_QUERY_LENGTH = 8
STATIC_CACHE_RESET_LENGTH = 8
STATIC_CACHE_RESPONSE_LENGTH = 8
STATIC_IPV4_PREFIX_LENGTH = 20
STATIC_IPV6_PREFIX_LENGTH = 32
STATIC_END_OF_DATA_V0_LENGTH = 12
STATIC_END_OF_DATA_V1_LENGTH = 24

RTR_VERSION = {0: '0',
               1: '1'}

PDU_TYPE = {0: 'Serial Notify',
            1: 'Serial Query',
            2: 'Reset Query',
            3: 'Cache Response',
            4: 'IPv4 Prefix',
            6: 'IPv6 Prefix',
            7: 'End of Data',
            8: 'Cache Reset',
            9: 'Router Key',
            10: 'Error Report',
            255: 'Reserved'}

ERROR_LIST = {0: 'Corrupt Data',
              1: 'Internal Error',
              2: 'No data Available',
              3: 'Invalid Request',
              4: 'Unsupported Protocol Version',
              5: 'Unsupported PDU Type',
              6: 'Withdrawal of Unknown Record',
              7: 'Duplicate Announcement Received',
              8: 'Unexpected Protocol Version'}


class RTRSerialNotify(Packet):

    '''

    Serial Notify packet from section 5.2
    https://tools.ietf.org/html/rfc6810#section-5.2

    '''

    name = 'Serial Notify'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 0, PDU_TYPE),
                   ShortField('session_id', 0),
                   IntField('length', STATIC_SERIAL_NOTIFY_LENGTH),
                   IntField('serial_number', 0)]


class RTRSerialQuery(Packet):

    '''

    Serial Query packet from section 5.3
    https://tools.ietf.org/html/rfc6810#section-5.3

    '''
    name = 'Serial Query'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 1, PDU_TYPE),
                   ShortField('session_id', 0),
                   IntField('length', STATIC_SERIAL_QUERY_LENGTH),
                   IntField('serial_number', 0)]


class RTRResetQuery(Packet):

    '''

    Reset Query packet from section 5.4
    https://tools.ietf.org/html/rfc6810#section-5.4

    '''
    name = 'Reset Query'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 2, PDU_TYPE),
                   ShortField('reserved', 0),
                   IntField('length', STATIC_RESET_QUERY_LENGTH)]


class RTRCacheResponse(Packet):

    '''

    Cache Response packet from section 5.5
    https://tools.ietf.org/html/rfc6810#section-5.5

    '''
    name = 'Cache Response'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 3, PDU_TYPE),
                   ShortField('session_id', 0),
                   IntField('length', STATIC_CACHE_RESPONSE_LENGTH)]

    def guess_payload_class(self, payload):
        return RTR


class RTRIPv4Prefix(Packet):

    '''

    IPv4 Prefix packet from section 5.6
    https://tools.ietf.org/html/rfc6810#section-5.6

    '''
    name = 'IPv4 Prefix'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 4, PDU_TYPE),
                   ShortField('reserved', 0),
                   IntField('length', STATIC_IPV4_PREFIX_LENGTH),
                   ByteField('flags', 0),
                   ByteField('shortest_length', 0),
                   ByteField('longest_length', 0),
                   ByteField('zeros', 0),
                   IPField('prefix', '0.0.0.0'),
                   IntField('asn', 0)]

    def guess_payload_class(self, payload):
        return RTR


class RTRIPv6Prefix(Packet):

    '''

    IPv6 Prefix packet from section 5.7
    https://tools.ietf.org/html/rfc6810#section-5.7

    '''
    name = 'IPv6 Prefix'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 6, PDU_TYPE),
                   ShortField('reserved', 0),
                   IntField('length', STATIC_IPV6_PREFIX_LENGTH),
                   ByteField('flags', 0),
                   ByteField('shortest_length', 0),
                   ByteField('longest_length', 0),
                   ByteField('zeros', 0),
                   IP6Field("prefix", "::"),
                   IntField('asn', 0)]

    def guess_payload_class(self, payload):
        return RTR


class RTREndofDatav0(Packet):

    '''

    End of Data packet from version 0 standard section 5.8
    https://tools.ietf.org/html/rfc6810#section-5.8

    '''
    name = 'End of Data - version 0'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 7, PDU_TYPE),
                   ShortField('session_id', 0),
                   IntField('length', STATIC_END_OF_DATA_V0_LENGTH),
                   IntField('serial_number', 0)]


class RTREndofDatav1(Packet):

    '''

    End of Data packet from version 1 standard section 5.8
    https://tools.ietf.org/html/rfc8210#section-5.8

    '''
    name = 'End of Data - version 1'
    fields_desc = [ByteEnumField('rtr_version', 1, RTR_VERSION),
                   ByteEnumField('pdu_type', 7, PDU_TYPE),
                   ShortField('session_id', 0),
                   IntField('length', STATIC_END_OF_DATA_V1_LENGTH),
                   IntField('serial_number', 0),
                   IntField('refresh_interval', 0),
                   IntField('retry_interval', 0),
                   IntField('expire_interval', 0)]


class RTRCacheReset(Packet):

    '''

    Cache Reset packet from section 5.9
    https://tools.ietf.org/html/rfc6810#section-5.9

    '''
    name = 'Reset Query'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 8, PDU_TYPE),
                   ShortField('reserved', 0),
                   IntField('length', STATIC_CACHE_RESET_LENGTH)]


class RTRRouterKey(Packet):

    '''

    Router Key packet from version 1 standard section 5.10
    https://tools.ietf.org/html/rfc8210#section-5.10

    '''
    name = 'Router Key'
    fields_desc = [ByteEnumField('rtr_version', 1, RTR_VERSION),
                   ByteEnumField('pdu_type', 9, PDU_TYPE),
                   ByteField('flags', 0),
                   ByteField('zeros', 0),
                   IntField('length', None),
                   StrFixedLenField('subject_key_identifier', '', 20),
                   IntField('asn', 0),
                   StrLenField('subject_PKI', '',
                               length_from=lambda x: x.length - 32)]

    def post_build(self, pkt, pay):
        temp_len = len(pkt) + 2
        if not self.length:
            pkt = pkt[:2] + struct.pack('!I', temp_len) + pkt[6:]
        return pkt + pay


class RTRErrorReport(Packet):

    '''

    Error Report packet from section 5.10
    https://tools.ietf.org/html/rfc6810#section-5.10

    '''
    name = 'Error Report'
    fields_desc = [ByteEnumField('rtr_version', 0, RTR_VERSION),
                   ByteEnumField('pdu_type', 10, PDU_TYPE),
                   ShortEnumField('error_code', 0, ERROR_LIST),
                   IntField('length', None),
                   FieldLenField('length_of_encaps_PDU',
                                 None, fmt='!I', length_of='erroneous_PDU'),
                   StrLenField('erroneous_PDU', '',
                               length_from=lambda x: x.length_of_encaps_PDU),
                   FieldLenField('length_of_error_text', None, fmt='!I',
                                 length_of='error_text'),
                   StrLenField('error_text', '',
                               length_from=lambda x: x.length_of_error_text)]

    def post_build(self, pkt, pay):
        temp_len = len(pkt) + 2
        if not self.length:
            pkt = pkt[:2] + struct.pack('!I', temp_len) + pkt[6:]
        return pkt + pay


PDU_CLASS_VERSION_0 = {0: RTRSerialNotify,
                       1: RTRSerialQuery,
                       2: RTRResetQuery,
                       3: RTRCacheResponse,
                       4: RTRIPv4Prefix,
                       6: RTRIPv6Prefix,
                       7: RTREndofDatav0,
                       8: RTRCacheReset,
                       10: RTRErrorReport}

PDU_CLASS_VERSION_1 = {0: RTRSerialNotify,
                       1: RTRSerialQuery,
                       2: RTRResetQuery,
                       3: RTRCacheResponse,
                       4: RTRIPv4Prefix,
                       6: RTRIPv6Prefix,
                       7: RTREndofDatav1,
                       8: RTRCacheReset,
                       9: RTRRouterKey,
                       10: RTRErrorReport}


class RTR(Packet):

    '''
    Dummy RPKI to Router generic packet for pre-sorting the packet type
    eg. https://tools.ietf.org/html/rfc6810#section-5.2

    '''
    name = 'RTR dissector'

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        '''
          Attribution of correct type depending on version and pdu_type
        '''
        if _pkt and len(_pkt) >= 2:
            version = orb(_pkt[0])
            pdu_type = orb(_pkt[1])
            if version == 0:
                return PDU_CLASS_VERSION_0[pdu_type]
            elif version == 1:
                return PDU_CLASS_VERSION_1[pdu_type]
        return Raw


bind_layers(TCP, RTR, dport=323)  # real reserved port
bind_layers(TCP, RTR, sport=323)  # real reserved port
bind_layers(TCP, RTR, dport=8282)  # RIPE implementation default port
bind_layers(TCP, RTR, sport=8282)  # RIPE implementation default port
bind_layers(TCP, RTR, dport=2222)  # gortr implementation default port
bind_layers(TCP, RTR, sport=2222)  # gortr implementation default port

if __name__ == '__main__':
    from scapy.main import interact
    interact(mydict=globals(), mybanner='RPKI to Router')



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = HomePlugSG Layer
# scapy.contrib.status = loads


from scapy.packet import Packet, bind_layers
from scapy.fields import FieldLenField, StrFixedLenField, StrLenField

# Extends HomePlug AV and GP layer
from scapy.contrib.homeplugav import HomePlugAV, QualcommTypeList

#
#    Copyright (C) HomePlugSG Layer for Scapy by FlUxIuS (Sebastien Dudek)
#

# HomePlug GP extension for SG


HomePlugSGTypes = {0xA400: "VS_UART_CMD_Req",
                   0xA401: "VS_UART_CMD_Cnf"}


QualcommTypeList.update(HomePlugSGTypes)

# UART commands over HomePlugGP


class VS_UART_CMD_REQ(Packet):
    name = "VS_UART_CMD_REQ"
    fields_desc = [FieldLenField("UDataLen", None, count_of="UData", fmt="H"),
                   StrLenField("UData", "UartCommand\x00",
                               length_from=lambda pkt: pkt.UDataLen)]


class VS_UART_CMD_CNF(Packet):
    name = "VS_UART_CMD_CNF"
    fields_desc = [StrFixedLenField("reserved", b"\x00", 6),
                   FieldLenField("UDataLen", None, count_of="UData", fmt="H"),
                   StrLenField("UData", "UartCommand\x00",
                               length_from=lambda pkt: pkt.UDataLen)]


# END #

bind_layers(HomePlugAV, VS_UART_CMD_REQ, HPtype=0xA400)
bind_layers(HomePlugAV, VS_UART_CMD_CNF, HPtype=0xA401)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Clone of p0f v3 passive OS fingerprinting
"""

import re
import struct
import random

from scapy.data import KnowledgeBase, select_path
from scapy.config import conf
from scapy.compat import raw, orb
from scapy.packet import NoPayload
from scapy.layers.inet import IP, TCP, TCPOptions
from scapy.layers.http import HTTP, HTTPRequest, HTTPResponse
from scapy.layers.inet6 import IPv6
from scapy.volatile import RandByte, RandShort, RandString
from scapy.error import warning

_p0fpaths = ["/etc/p0f", "/usr/share/p0f", "/opt/local"]
conf.p0f_base = select_path(_p0fpaths, "p0f.fp")

MIN_TCP4 = 40  # Min size of IPv4/TCP headers
MIN_TCP6 = 60  # Min size of IPv6/TCP headers
MAX_DIST = 35  # Maximum TTL distance for non-fuzzy signature matching

WIN_TYPE_NORMAL = 0  # Literal value
WIN_TYPE_ANY = 1  # Wildcard
WIN_TYPE_MOD = 2  # Modulo check
WIN_TYPE_MSS = 3  # Window size MSS multiplier
WIN_TYPE_MTU = 4  # Window size MTU multiplier

# Convert TCP option num to p0f (nop is handled separately)
tcp_options_p0f = {
    2: "mss",  # maximum segment size
    3: "ws",  # window scaling
    4: "sok",  # selective ACK permitted
    5: "sack",  # selective ACK (should not be seen)
    8: "ts",  # timestamp
}


# Signatures
class TCP_Signature(object):
    __slots__ = ["olayout", "quirks", "ip_opt_len", "ip_ver", "ttl",
                 "mss", "win", "win_type", "wscale", "pay_class", "ts1"]

    def __init__(self, olayout, quirks, ip_opt_len, ip_ver, ttl,
                 mss, win, win_type, wscale, pay_class, ts1):
        self.olayout = olayout
        self.quirks = quirks
        self.ip_opt_len = ip_opt_len
        self.ip_ver = ip_ver
        self.ttl = ttl
        self.mss = mss
        self.win = win
        self.win_type = win_type  # None for packet signatures
        self.wscale = wscale
        self.pay_class = pay_class
        self.ts1 = ts1  # None for base signatures

    @classmethod
    def from_packet(cls, pkt):
        """
        Receives a TCP packet (assuming it's valid), and returns
        a TCP_Signature object
        """
        ip_ver = pkt.version
        quirks = set()

        def addq(name):
            quirks.add(name)

        # IPv4/IPv6 parsing
        if ip_ver == 4:
            ttl = pkt.ttl
            ip_opt_len = (pkt.ihl * 4) - 20
            if pkt.tos & (0x01 | 0x02):
                addq("ecn")
            if pkt.flags.evil:
                addq("0+")
            if pkt.flags.DF:
                addq("df")
                if pkt.id:
                    addq("id+")
            elif pkt.id == 0:
                addq("id-")
        else:
            ttl = pkt.hlim
            ip_opt_len = 0
            if pkt.fl:
                addq("flow")
            if pkt.tc & (0x01 | 0x02):
                addq("ecn")

        # TCP parsing
        tcp = pkt[TCP]
        win = tcp.window
        if tcp.flags & (0x40 | 0x80 | 0x01):
            addq("ecn")
        if tcp.seq == 0:
            addq("seq-")
        if tcp.flags.A:
            if tcp.ack == 0:
                addq("ack-")
        elif tcp.ack:
            addq("ack+")
        if tcp.flags.U:
            addq("urgf+")
        elif tcp.urgptr:
            addq("uptr+")
        if tcp.flags.P:
            addq("pushf+")

        pay_class = 1 if tcp.payload else 0

        # Manual TCP options parsing
        mss = 0
        wscale = 0
        ts1 = 0
        olayout = ""
        optlen = (tcp.dataofs << 2) - 20
        x = raw(tcp)[-optlen:]  # raw bytes of TCP options
        while x:
            onum = orb(x[0])
            if onum == 0:
                x = x[1:]
                olayout += "eol+%i," % len(x)
                if x.strip(b"\x00"):  # non-zero past EOL
                    addq("opt+")
                break
            if onum == 1:
                x = x[1:]
                olayout += "nop,"
                continue
            try:
                olen = orb(x[1])
            except IndexError:  # no room for length field
                addq("bad")
                break
            oval = x[2:olen]
            if onum in tcp_options_p0f:
                ofmt = TCPOptions[0][onum][1]
                olayout += "%s," % tcp_options_p0f[onum]
                optsize = 2 + struct.calcsize(ofmt) if ofmt else 2  # total len
                if len(x) < optsize:  # option would end past end of header
                    addq("bad")
                    break

                if onum == 5:
                    if olen < 10 or olen > 34:  # SACK length out of range
                        addq("bad")
                        break
                else:
                    if olen != optsize:  # length field doesn't fit option type
                        addq("bad")
                        break
                    if ofmt:
                        oval = struct.unpack(ofmt, oval)
                        if len(oval) == 1:
                            oval = oval[0]
                    if onum == 2:
                        mss = oval
                    elif onum == 3:
                        wscale = oval
                        if wscale > 14:
                            addq("exws")
                    elif onum == 8:
                        ts1 = oval[0]
                        if not ts1:
                            addq("ts1-")
                        if oval[1] and (tcp.flags.S and not tcp.flags.A):
                            addq("ts2+")
            else:  # Unknown option, presumably with specified size
                if olen < 2 or olen > 40 or olen > len(x):
                    addq("bad")
                    break
            x = x[olen:]
        olayout = olayout[:-1]

        return cls(olayout, quirks, ip_opt_len, ip_ver, ttl, mss, win, None, wscale, pay_class, ts1)  # noqa: E501

    @classmethod
    def from_raw_sig(cls, sig_line):
        """
        Parses a TCP sig line and returns a tuple consisting of a
        TCP_Signature object and bad_ttl as bool
        """
        ver, ttl, olen, mss, wsize, olayout, quirks, pclass = lparse(sig_line, 8)  # noqa: E501
        wsize, _, scale = wsize.partition(",")

        ip_ver = -1 if ver == "*" else int(ver)
        ttl, bad_ttl = (int(ttl[:-1]), True) if ttl[-1] == "-" else (int(ttl), False)  # noqa: E501
        ip_opt_len = int(olen)
        mss = -1 if mss == "*" else int(mss)
        if wsize == "*":
            win, win_type = (0, WIN_TYPE_ANY)
        elif wsize[:3] == "mss":
            win, win_type = (int(wsize[4:]), WIN_TYPE_MSS)
        elif wsize[0] == "%":
            win, win_type = (int(wsize[1:]), WIN_TYPE_MOD)
        elif wsize[:3] == "mtu":
            win, win_type = (int(wsize[4:]), WIN_TYPE_MTU)
        else:
            win, win_type = (int(wsize), WIN_TYPE_NORMAL)
        wscale = -1 if scale == "*" else int(scale)
        if quirks:
            quirks = frozenset(q for q in quirks.split(","))
        else:
            quirks = frozenset()
        pay_class = -1 if pclass == "*" else int(pclass == "+")

        sig = cls(olayout, quirks, ip_opt_len, ip_ver, ttl, mss, win, win_type, wscale, pay_class, None)  # noqa: E501
        return sig, bad_ttl

    def __str__(self):
        quirks = ",".join(q for q in self.quirks)
        fmt = "%i:%i+%i:%i:%i:%i,%i:%s:%s:%i"
        s = fmt % (self.ip_ver, self.ttl, guess_dist(self.ttl),
                   self.ip_opt_len, self.mss, self.win, self.wscale,
                   self.olayout, quirks, self.pay_class)
        return s


class HTTP_Signature(object):
    __slots__ = ["http_ver", "hdr", "hdr_set", "habsent", "sw"]

    def __init__(self, http_ver, hdr, hdr_set, habsent, sw):
        self.http_ver = http_ver
        self.hdr = hdr
        self.hdr_set = hdr_set
        self.habsent = habsent  # None for packet signatures
        self.sw = sw

    @classmethod
    def from_packet(cls, pkt):
        """
        Receives an HTTP packet (assuming it's valid), and returns
        a HTTP_Signature object
        """
        http_payload = raw(pkt[TCP].payload)

        crlfcrlf = b"\r\n\r\n"
        crlfcrlfIndex = http_payload.find(crlfcrlf)
        if crlfcrlfIndex != -1:
            headers = http_payload[:crlfcrlfIndex + len(crlfcrlf)]
        else:
            headers = http_payload
        headers = headers.decode()  # XXX: Check if this could fail
        first_line, headers = headers.split("\r\n", 1)

        if "1.0" in first_line:
            http_ver = 0
        elif "1.1" in first_line:
            http_ver = 1
        else:
            raise ValueError("HTTP version is not 1.0/1.1")

        sw = ""
        headers_found = []
        hdr_set = set()
        for header_line in headers.split("\r\n"):
            name, _, value = header_line.partition(":")
            if value:
                value = value.strip()
                headers_found.append((name, value))
                hdr_set.add(name)
                if name in ("User-Agent", "Server"):
                    sw = value
        hdr = tuple(headers_found)
        return cls(http_ver, hdr, hdr_set, None, sw)

    @classmethod
    def from_raw_sig(cls, sig_line):
        """
        Parses an HTTP sig line and returns a HTTP_Signature object
        """
        ver, horder, habsent, expsw = lparse(sig_line, 4)
        http_ver = -1 if ver == "*" else int(ver)

        # horder parsing - split by commas that aren't in []
        new_horder = []
        for header in re.split(r",(?![^\[]*\])", horder):
            name, _, value = header.partition("=")
            if name[0] == "?":  # Optional header
                new_horder.append((name[1:], value[1:-1], True))
            else:
                new_horder.append((name, value[1:-1], False))
        hdr = tuple(new_horder)
        hdr_set = frozenset(header[0] for header in hdr if not header[2])
        habsent = frozenset(habsent.split(","))
        return cls(http_ver, hdr, hdr_set, habsent, expsw)

    def __str__(self):
        # values that depend on the context are not included in the string
        skipval = ("Host", "User-Agent", "Date", "Content-Type", "Server")
        hdr = ",".join(n if n in skipval else "%s=[%s]" % (n, v) for n, v in self.hdr)  # noqa: E501
        fmt = "%i:%s::%s"
        s = fmt % (self.http_ver, hdr, self.sw)
        return s


# Records
class MTU_Record(object):
    __slots__ = ["label_id", "mtu"]

    def __init__(self, label_id, sig_line):
        self.label_id = label_id
        self.mtu = int(sig_line)


class TCP_Record(object):
    __slots__ = ["label_id", "bad_ttl", "sig"]

    def __init__(self, label_id, sig_line):
        self.label_id = label_id
        sig, bad_ttl = TCP_Signature.from_raw_sig(sig_line)
        self.bad_ttl = bad_ttl
        self.sig = sig


class HTTP_Record(object):
    __slots__ = ["label_id", "sig"]

    def __init__(self, label_id, sig_line):
        self.label_id = label_id
        self.sig = HTTP_Signature.from_raw_sig(sig_line)


class p0fKnowledgeBase(KnowledgeBase):
    """
    .. code::

        self.base = {
            "mtu" (str): [sig(tuple), ...]
            "tcp"/"http" (str): {
                direction (str): [sig(tuple), ...]
            }
        }
        self.labels = (label(tuple), ...)

    """
    def lazy_init(self):
        try:
            f = open(self.filename)
        except Exception:
            warning("Can't open base %s", self.filename)
            return

        self.base = {}
        self.labels = []
        self._parse_file(f)
        self.labels = tuple(self.labels)
        f.close()

    def _parse_file(self, file):
        """
        Parses p0f.fp file and stores the data with described structures.
        """
        label_id = -1

        for line in file:
            if line[0] in (";", "\n"):
                continue
            line = line.strip()

            if line[0] == "[":
                section, direction = lparse(line[1:-1], 2)
                if section == "mtu":
                    self.base[section] = []
                    curr_records = self.base[section]
                else:
                    if section not in self.base:
                        self.base[section] = {direction: []}
                    elif direction not in self.base[section]:
                        self.base[section][direction] = []
                    curr_records = self.base[section][direction]
            else:
                param, _, val = line.partition(" = ")
                param = param.strip()

                if param == "sig":
                    if section == "mtu":
                        record_class = MTU_Record
                    elif section == "tcp":
                        record_class = TCP_Record
                    elif section == "http":
                        record_class = HTTP_Record
                    curr_records.append(record_class(label_id, val))

                elif param == "label":
                    label_id += 1
                    if section == "mtu":
                        self.labels.append(val)
                        continue
                    # label = type:class:name:flavor
                    t, c, name, flavor = lparse(val, 4)
                    self.labels.append((t, c, name, flavor))

                elif param == "sys":
                    sys_names = tuple(name for name in val.split(","))
                    self.labels[label_id] += (sys_names,)

    def get_sigs_by_os(self, direction, osgenre, osdetails=None):
        """Get TCP signatures that match an OS genre and details (if specified).
        If osdetails isn't specified, then we pick all signatures
        that match osgenre.

        Examples:
            >>> p0fdb.get_sigs_by_os("request", "Linux", "2.6")
            >>> p0fdb.get_sigs_by_os("response", "Windows", "8")
            >>> p0fdb.get_sigs_by_os("request", "FreeBSD")
        """
        sigs = []
        for tcp_record in self.base["tcp"][direction]:
            label = self.labels[tcp_record.label_id]
            name, flavor = label[2], label[3]
            if osgenre and osgenre == name:
                if osdetails:
                    if osdetails in flavor:
                        sigs.append(tcp_record.sig)
                else:
                    sigs.append(tcp_record.sig)
        return sigs

    def tcp_find_match(self, ts, direction):
        """
        Finds the best match for the given signature and direction.
        If a match is found, returns a tuple consisting of:
        - label: the matched label
        - dist: guessed distance from the packet source
        - fuzzy: whether the match is fuzzy
        Returns None if no match was found
        """
        win_multi, use_mtu = detect_win_multi(ts)

        gmatch = None  # generic match
        fmatch = None  # fuzzy match
        for tcp_record in self.base["tcp"][direction]:
            rs = tcp_record.sig

            fuzzy = False
            ref_quirks = rs.quirks

            if rs.olayout != ts.olayout:
                continue

            if rs.ip_ver == -1:
                ref_quirks -= {"flow"} if ts.ip_ver == 4 else {"df", "id+", "id-"}  # noqa: E501

            if ref_quirks != ts.quirks:
                deleted = (ref_quirks ^ ts.quirks) & ref_quirks
                added = (ref_quirks ^ ts.quirks) & ts.quirks

                if (fmatch or (deleted - {"df", "id+"}) or (added - {"id-", "ecn"})):  # noqa: E501
                    continue
                fuzzy = True

            if rs.ip_opt_len != ts.ip_opt_len:
                continue
            if tcp_record.bad_ttl:
                if rs.ttl < ts.ttl:
                    continue
            else:
                if rs.ttl < ts.ttl or rs.ttl - ts.ttl > MAX_DIST:
                    fuzzy = True

            if ((rs.mss != -1 and rs.mss != ts.mss) or
               (rs.wscale != -1 and rs.wscale != ts.wscale) or
               (rs.pay_class != -1 and rs.pay_class != ts.pay_class)):
                continue

            if rs.win_type == WIN_TYPE_NORMAL:
                if rs.win != ts.win:
                    continue
            elif rs.win_type == WIN_TYPE_MOD:
                if ts.win % rs.win:
                    continue
            elif rs.win_type == WIN_TYPE_MSS:
                if (use_mtu or rs.win != win_multi):
                    continue
            elif rs.win_type == WIN_TYPE_MTU:
                if (not use_mtu or rs.win != win_multi):
                    continue

            # Got a match? If not fuzzy, return. If fuzzy, keep looking.
            label = self.labels[tcp_record.label_id]
            match = (label, rs.ttl - ts.ttl, fuzzy)
            if not fuzzy:
                if label[0] == "s":
                    return match
                elif not gmatch:
                    gmatch = match
            elif not fmatch:
                fmatch = match

        if gmatch:
            return gmatch
        if fmatch:
            return fmatch
        return None

    def http_find_match(self, ts, direction):
        """
        Finds the best match for the given signature and direction.
        If a match is found, returns a tuple consisting of:
        - label: the matched label
        - dishonest: whether the software was detected as dishonest
        Returns None if no match was found
        """
        gmatch = None  # generic match
        for http_record in self.base["http"][direction]:
            rs = http_record.sig

            if rs.http_ver != -1 and rs.http_ver != ts.http_ver:
                continue

            # Check that all non-optional headers appear in the packet
            if not (ts.hdr_set & rs.hdr_set) == rs.hdr_set:
                continue

            # Check that no forbidden headers appear in the packet.
            if len(rs.habsent & ts.hdr_set) > 0:
                continue

            def headers_correl():
                phi = 0  # Packet HTTP header index
                hdr_len = len(ts.hdr)

                # Confirm the ordering and values of headers
                # (this is relatively slow, hence the if statements above).
                # The algorithm is derived from the original p0f/fp_http.c
                for kh in rs.hdr:
                    orig_phi = phi
                    while (phi < hdr_len and
                           kh[0] != ts.hdr[phi][0]):
                        phi += 1

                    if phi == hdr_len:
                        if not kh[2]:
                            return False

                        for ph in ts.hdr:
                            if kh[0] == ph[0]:
                                return False

                        phi = orig_phi
                        continue

                    if kh[1] not in ts.hdr[phi][1]:
                        return False
                    phi += 1
                return True

            if not headers_correl():
                continue

            # Got a match
            label = self.labels[http_record.label_id]
            dishonest = rs.sw and ts.sw and rs.sw not in ts.sw
            match = (label, dishonest)
            if label[0] == "s":
                return match
            elif not gmatch:
                gmatch = match
        return gmatch if gmatch else None

    def mtu_find_match(self, mtu):
        """
        Finds a match for the given MTU.
        If a match is found, returns the label string.
        Returns None if no match was found
        """
        for mtu_record in self.base["mtu"]:
            if mtu == mtu_record.mtu:
                return self.labels[mtu_record.label_id]
        return None


p0fdb = p0fKnowledgeBase(conf.p0f_base)


def guess_dist(ttl):
    for ottl in (32, 64, 128, 255):
        if ttl <= ottl:
            return ottl - ttl


def lparse(line, n, delimiter=":", default=""):
    """
    Parsing of 'a:b:c:d:e' lines
    """
    a = line.split(delimiter)[:n]
    for elt in a:
        yield elt
    for _ in range(n - len(a)):
        yield default


def validate_packet(pkt):
    """
    Validate that the packet is an IPv4/IPv6 and TCP packet.
    If the packet is valid, a copy is returned. If not, TypeError is raised.
    """
    pkt = pkt.copy()
    valid = pkt.haslayer(TCP) and (pkt.haslayer(IP) or pkt.haslayer(IPv6))
    if not valid:
        raise TypeError("Not a TCP/IP packet")
    return pkt


def detect_win_multi(ts):
    """
    Figure out if window size is a multiplier of MSS or MTU.
    Receives a TCP signature and returns the multiplier and
    whether mtu should be used
    """
    mss = ts.mss
    win = ts.win
    if not win or mss < 100:
        return -1, False

    options = [
        (mss, False),
        (1500 - MIN_TCP4, False),
        (1500 - MIN_TCP4 - 12, False),
        (mss + MIN_TCP4, True),
        (1500, True)
    ]
    if ts.ts1:
        options.append((mss - 12, False))
    if ts.ip_ver == 6:
        options.append((1500 - MIN_TCP6, False))
        options.append((1500 - MIN_TCP6 - 12, False))
        options.append((mss + MIN_TCP6, True))

    for div, use_mtu in options:
        if not win % div:
            return win / div, use_mtu
    return -1, False


def packet2p0f(pkt):
    """
    Returns a p0f signature of the packet, and the direction.
    Raises TypeError if the packet isn't valid for p0f
    """
    pkt = validate_packet(pkt)
    pkt = pkt.__class__(raw(pkt))

    if pkt[TCP].flags.S:
        if pkt[TCP].flags.A:
            direction = "response"
        else:
            direction = "request"
        sig = TCP_Signature.from_packet(pkt)

    elif pkt[TCP].payload:
        # XXX: guess_payload_class doesn't use any class related attributes
        pclass = HTTP().guess_payload_class(raw(pkt[TCP].payload))
        if pclass == HTTPRequest:
            direction = "request"
        elif pclass == HTTPResponse:
            direction = "response"
        else:
            raise TypeError("Not an HTTP payload")
        sig = HTTP_Signature.from_packet(pkt)
    else:
        raise TypeError("Not a SYN, SYN/ACK, or HTTP packet")
    return sig, direction


def fingerprint_mtu(pkt):
    """
    Fingerprints the MTU based on the maximum segment size specified
    in TCP options.
    If a match was found, returns the label. If not returns None
    """
    pkt = validate_packet(pkt)
    mss = 0
    for name, value in pkt.payload.options:
        if name == "MSS":
            mss = value

    if not mss:
        return None

    mtu = (mss + MIN_TCP4) if pkt.version == 4 else (mss + MIN_TCP6)

    if not p0fdb.get_base():
        warning("p0f base empty.")
        return None

    return p0fdb.mtu_find_match(mtu)


def p0f(pkt):
    sig, direction = packet2p0f(pkt)
    if not p0fdb.get_base():
        warning("p0f base empty.")
        return None

    if isinstance(sig, TCP_Signature):
        return p0fdb.tcp_find_match(sig, direction)
    else:
        return p0fdb.http_find_match(sig, direction)


def prnp0f(pkt):
    """Calls p0f and prints a user-friendly output"""
    try:
        r = p0f(pkt)
    except Exception:
        return

    sig, direction = packet2p0f(pkt)
    is_tcp_sig = isinstance(sig, TCP_Signature)
    to_server = direction == "request"

    if is_tcp_sig:
        pkt_type = "SYN" if to_server else "SYN+ACK"
    else:
        pkt_type = "HTTP Request" if to_server else "HTTP Response"

    res = pkt.sprintf(".-[ %IP.src%:%TCP.sport% -> %IP.dst%:%TCP.dport% (" + pkt_type + ") ]-\n|\n")  # noqa: E501
    fields = []

    def add_field(name, value):
        fields.append("| %-8s = %s\n" % (name, value))

    cli_or_svr = "Client" if to_server else "Server"
    add_field(cli_or_svr, pkt.sprintf("%IP.src%:%TCP.sport%"))

    if r:
        label = r[0]
        app_or_os = "App" if label[1] == "!" else "OS"
        add_field(app_or_os, label[2] + " " + label[3])
        if len(label) == 5:  # label includes sys
            add_field("Sys", ", ".join(name for name in label[4]))
        if is_tcp_sig:
            add_field("Distance", r[1])
    else:
        app_or_os = "OS" if is_tcp_sig else "App"
        add_field(app_or_os, "UNKNOWN")

    add_field("Raw sig", str(sig))

    res += "".join(fields)
    res += "`____\n"
    print(res)


def p0f_impersonate(pkt, osgenre=None, osdetails=None, signature=None,
                    extrahops=0, mtu=1500, uptime=None):
    """
    Modifies pkt so that p0f will think it has been sent by a
    specific OS. Either osgenre or signature is required to impersonate.
    If signature is specified (as a raw string), we use the signature.
    signature format::

        "ip_ver:ttl:ip_opt_len:mss:window,wscale:opt_layout:quirks:pay_class"

    If osgenre is specified, we randomly pick a signature with a label
    that matches osgenre (and osdetails, if specified).
    Note: osgenre is case sensitive ("linux" -> "Linux" etc.), and osdetails
    is a substring of a label flavor ("7", "8" and "7 or 8" will
    all match the label "s:win:Windows:7 or 8")

    For now, only TCP SYN/SYN+ACK packets are supported.
    """
    pkt = validate_packet(pkt)

    if not osgenre and not signature:
        raise ValueError("osgenre or signature is required to impersonate!")

    tcp = pkt[TCP]
    tcp_type = tcp.flags & (0x02 | 0x10)  # SYN / SYN+ACK

    if signature:
        if isinstance(signature, str):
            sig, _ = TCP_Signature.from_raw_sig(signature)
        else:
            raise TypeError("Unsupported signature type")
    else:
        if not p0fdb.get_base():
            sigs = []
        else:
            direction = "request" if tcp_type == 0x02 else "response"
            sigs = p0fdb.get_sigs_by_os(direction, osgenre, osdetails)

        # If IPv6 packet, remove IPv4-only signatures and vice versa
        sigs = [s for s in sigs if s.ip_ver == -1 or s.ip_ver == pkt.version]
        if not sigs:
            raise ValueError("No match in the p0f database")
        sig = random.choice(sigs)

    if sig.ip_ver != -1 and pkt.version != sig.ip_ver:
        raise ValueError("Can't convert between IPv4 and IPv6")

    quirks = sig.quirks

    if pkt.version == 4:
        pkt.ttl = sig.ttl - extrahops
        if sig.ip_opt_len != 0:
            # FIXME: Non-zero IPv4 options not handled
            warning("Unhandled IPv4 option field")
        else:
            pkt.options = []

        if "df" in quirks:
            pkt.flags |= 0x02  # set DF flag
            if "id+" in quirks:
                if pkt.id == 0:
                    pkt.id = random.randint(1, 2**16 - 1)
            else:
                pkt.id = 0
        else:
            pkt.flags &= ~(0x02)  # DF flag not set
            if "id-" in quirks:
                pkt.id = 0
            elif pkt.id == 0:
                pkt.id = random.randint(1, 2**16 - 1)
        if "ecn" in quirks:
            pkt.tos |= random.randint(0x01, 0x03)
        pkt.flags = pkt.flags | 0x04 if "0+" in quirks else pkt.flags & ~(0x04)
    else:
        pkt.hlim = sig.ttl - extrahops
        if "flow" in quirks:
            pkt.fl = random.randint(1, 2**20 - 1)
        if "ecn" in quirks:
            pkt.tc |= random.randint(0x01, 0x03)

    # Take the options already set as "hints" to use in the new packet if we
    # can. we'll use the already-set values if they're valid integers.
    def int_only(val):
        return val if isinstance(val, int) else None
    orig_opts = dict(tcp.options)
    mss_hint = int_only(orig_opts.get("MSS"))
    ws_hint = int_only(orig_opts.get("WScale"))
    ts_hint = [int_only(o) for o in orig_opts.get("Timestamp", (None, None))]

    options = []
    for opt in sig.olayout.split(","):
        if opt == "mss":
            # MSS might have a maximum size because of WIN_TYPE_MSS
            if sig.win_type == WIN_TYPE_MSS:
                maxmss = (2**16 - 1) // sig.win
            else:
                maxmss = (2**16 - 1)

            if sig.mss == -1:  # wildcard mss
                if mss_hint and 0 <= mss_hint <= maxmss:
                    options.append(("MSS", mss_hint))
                else:  # invalid hint, generate new value
                    options.append(("MSS", random.randint(100, maxmss)))
            else:
                options.append(("MSS", sig.mss))

        elif opt == "ws":
            if sig.wscale == -1:  # wildcard wscale
                maxws = 2**8
                if "exws" in quirks:  # wscale > 14
                    if ws_hint and 14 < ws_hint < maxws:
                        options.append(("WScale", ws_hint))
                    else:  # invalid hint, generate new value > 14
                        options.append(("WScale", random.randint(15, maxws - 1)))  # noqa: E501
                else:
                    if ws_hint and 0 <= ws_hint < maxws:
                        options.append(("WScale", ws_hint))
                    else:  # invalid hint, generate new value
                        options.append(("WScale", RandByte()))
            else:
                options.append(("WScale", sig.wscale))

        elif opt == "ts":
            ts1, ts2 = ts_hint

            if "ts1-" in quirks:  # own timestamp specified as zero
                ts1 = 0
            elif uptime is not None:  # if specified uptime, override
                ts1 = uptime
            elif ts1 is None or not (0 < ts1 < 2**32):  # invalid hint
                ts1 = random.randint(120, 100 * 60 * 60 * 24 * 365)

            # non-zero peer timestamp on initial SYN
            if "ts2+" in quirks and tcp_type == 0x02:
                if ts2 is None or not (0 < ts2 < 2**32):  # invalid hint
                    ts2 = random.randint(1, 2**32 - 1)
            else:
                ts2 = 0
            options.append(("Timestamp", (ts1, ts2)))

        elif opt == "nop":
            options.append(("NOP", None))
        elif opt == "sok":
            options.append(("SAckOK", ""))
        elif opt[:3] == "eol":
            options.append(("EOL", None))
            # FIXME: opt+ quirk not handled
            if "opt+" in quirks:
                warning("Unhandled opt+ quirk")
        elif opt == "sack":
            # Randomize SAck value in range of 10 <= val <= 34
            sack_len = random.choice([10, 18, 26, 34]) - 2
            optstruct = "!%iI" % (sack_len // 4)
            rand_val = RandString(struct.calcsize(optstruct))._fix()
            options.append(("SAck", struct.unpack(optstruct, rand_val)))
        else:
            warning("Unhandled TCP option %s", opt)
        tcp.options = options

    if sig.win_type == WIN_TYPE_NORMAL:
        tcp.window = sig.win
    elif sig.win_type == WIN_TYPE_MSS:
        mss = [x for x in options if x[0] == "MSS"]
        if not mss:
            raise ValueError("TCP window value requires MSS, and MSS option not set")  # noqa: E501
        tcp.window = mss[0][1] * sig.win
    elif sig.win_type == WIN_TYPE_MOD:
        tcp.window = sig.win * random.randint(1, (2**16 - 1) // sig.win)
    elif sig.win_type == WIN_TYPE_MTU:
        tcp.window = mtu * sig.win
    elif sig.win_type == WIN_TYPE_ANY:
        tcp.window = RandShort()
    else:
        warning("Unhandled window size specification")

    if "seq-" in quirks:
        tcp.seq = 0
    elif tcp.seq == 0:
        tcp.seq = random.randint(1, 2**32 - 1)

    if "ack+" in quirks:
        tcp.flags &= ~(0x10)  # ACK flag not set
        if tcp.ack == 0:
            tcp.ack = random.randint(1, 2**32 - 1)
    elif "ack-" in quirks:
        tcp.flags |= 0x10  # ACK flag set
        tcp.ack = 0

    if "uptr+" in quirks:
        tcp.flags &= ~(0x020)  # URG flag not set
        if tcp.urgptr == 0:
            tcp.urgptr = random.randint(1, 2**16 - 1)
    elif "urgf+" in quirks:
        tcp.flags |= 0x020  # URG flag used

    tcp.flags = tcp.flags | 0x08 if "pushf+" in quirks else tcp.flags & ~(0x08)

    if sig.pay_class:  # signature has payload
        if not tcp.payload:
            pkt /= conf.raw_layer(load=RandString(random.randint(1, 10)))
    else:
        tcp.payload = NoPayload()

    return pkt



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Thomas Tannhaeuser <hecke@naberius.de>

# scapy.contrib.status = skip


# Package of contrib SCADA modules.


"""contains packages related to SCADA protocol layers."""

from scapy.contrib.scada.iec104 import *  # noqa F403,F401



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = Protocol Independent Multicast (PIM)
# scapy.contrib.status = loads
"""
References:
  - https://tools.ietf.org/html/rfc4601
  - https://www.iana.org/assignments/pim-parameters/pim-parameters.xhtml
"""
import struct
from scapy.packet import Packet, bind_layers
from scapy.fields import BitFieldLenField, BitField, BitEnumField, ByteField, \
    ShortField, XShortField, IPField, IP6Field, PacketListField, \
    IntField, FieldLenField, BoundStrLenField, MultipleTypeField
from scapy.layers.inet import IP
from scapy.layers.inet6 import IPv6, in6_chksum, _IPv6ExtHdr
from scapy.utils import checksum
from scapy.compat import orb
from scapy.config import conf
from scapy.volatile import RandInt


PIM_TYPE = {
    0: "Hello",
    1: "Register",
    2: "Register-Stop",
    3: "Join/Prune",
    4: "Bootstrap",
    5: "Assert",
    6: "Graft",
    7: "Graft-Ack",
    8: "Candidate-RP-Advertisement"
}


class PIMv2Hdr(Packet):
    name = "Protocol Independent Multicast Version 2 Header"
    fields_desc = [BitField("version", 2, 4),
                   BitEnumField("type", 0, 4, PIM_TYPE),
                   ByteField("reserved", 0),
                   XShortField("chksum", None)]

    def post_build(self, p, pay):
        """
        Called implicitly before a packet is sent to compute and
         place PIM checksum.

        Parameters:
          self    The instantiation of an PIMv2Hdr class
          p       The PIMv2Hdr message in hex in network byte order
          pay     Additional payload for the PIMv2Hdr message
        """
        p += pay
        if self.chksum is None:
            if isinstance(self.underlayer, IP):
                ck = checksum(p)
                # ck = in4_chksum(103, self.underlayer, p)
                # According to RFC768 if the result checksum is 0, it should be set to 0xFFFF  # noqa: E501
                if ck == 0:
                    ck = 0xFFFF
                p = p[:2] + struct.pack("!H", ck) + p[4:]

            elif isinstance(self.underlayer, IPv6) or isinstance(self.underlayer, _IPv6ExtHdr):  # noqa: E501
                ck = in6_chksum(103, self.underlayer, p)  # noqa: E501
                # According to RFC2460 if the result checksum is 0, it should be set to 0xFFFF  # noqa: E501
                if ck == 0:
                    ck = 0xFFFF
                p = p[:2] + struct.pack("!H", ck) + p[4:]

        return p


def _guess_pim_tlv_class(h_classes, default_key, pkt, **kargs):
    cls = conf.raw_layer
    if len(pkt) >= 2:
        tlvtype = orb(pkt[1])
        cls = h_classes.get(tlvtype, default_key)
    return cls(pkt, **kargs)


class _PIMGenericTlvBase(Packet):
    fields_desc = [ByteField("type", 0),
                   FieldLenField("length", None, length_of="value", fmt="B"),
                   BoundStrLenField("value", "",
                                    length_from=lambda pkt: pkt.length)]

    def guess_payload_class(self, p):
        return conf.padding_layer

    def extract_padding(self, s):
        return "", s


##################################
# PIMv2 Hello
##################################
class _PIMv2GenericHello(_PIMGenericTlvBase):
    name = "PIMv2 Generic Hello"


def _guess_pimv2_hello_class(p, **kargs):
    return _guess_pim_tlv_class(PIMv2_HELLO_CLASSES, None, p, **kargs)


class _PIMv2HelloListField(PacketListField):
    def __init__(self):
        PacketListField.__init__(self, "option", [], _guess_pimv2_hello_class)


class PIMv2Hello(Packet):
    name = "PIMv2 Hello Options"
    fields_desc = [
        _PIMv2HelloListField()
    ]


class PIMv2HelloHoldtime(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : Holdtime"
    fields_desc = [
        ShortField("type", 1),
        FieldLenField("length", None, length_of="holdtime", fmt="!H"),
        ShortField("holdtime", 105)
    ]


class PIMv2HelloLANPruneDelayValue(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : LAN Prune Delay Value"
    fields_desc = [
        BitField("t", 0, 1),
        BitField("propagation_delay", 500, 15),
        ShortField("override_interval", 2500),
    ]


class PIMv2HelloLANPruneDelay(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : LAN Prune Delay"
    fields_desc = [
        ShortField("type", 2),
        FieldLenField("length", None, length_of="value", fmt="!H"),
        PacketListField("value", PIMv2HelloLANPruneDelayValue(),
                        PIMv2HelloLANPruneDelayValue,
                        length_from=lambda pkt: pkt.length)
    ]


class PIMv2HelloDRPriority(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : DR Priority"
    fields_desc = [
        ShortField("type", 19),
        FieldLenField("length", None, length_of="dr_priority", fmt="!H"),
        IntField("dr_priority", 1)
    ]


class PIMv2HelloGenerationID(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : Generation ID"
    fields_desc = [
        ShortField("type", 20),
        FieldLenField(
            "length", None, length_of="generation_id", fmt="!H"
        ),
        IntField("generation_id", RandInt())
    ]


class PIMv2HelloStateRefreshValue(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : State-Refresh Value"
    fields_desc = [ByteField("version", 1),
                   ByteField("interval", 0),
                   ShortField("reserved", 0)]


class PIMv2HelloStateRefresh(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : State-Refresh"
    fields_desc = [
        ShortField("type", 21),
        FieldLenField(
            "length", None, length_of="value", fmt="!H"
        ),
        PacketListField("value", PIMv2HelloStateRefreshValue(),
                        PIMv2HelloStateRefreshValue)
    ]


class PIMv2HelloAddrListValue(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : Address List Value"
    fields_desc = [
        ByteField("addr_family", 1),
        ByteField("encoding_type", 0),
        IP6Field("prefix", "::"),
    ]


class PIMv2HelloAddrList(_PIMv2GenericHello):
    name = "PIMv2 Hello Options : Address List"
    fields_desc = [
        ShortField("type", 24),
        FieldLenField(
            "length", None, length_of="value" , fmt="!H"
        ),
        PacketListField("value", PIMv2HelloAddrListValue(),
                        PIMv2HelloAddrListValue)
    ]


PIMv2_HELLO_CLASSES = {
    1: PIMv2HelloHoldtime,
    2: PIMv2HelloLANPruneDelay,
    19: PIMv2HelloDRPriority,
    20: PIMv2HelloGenerationID,
    21: PIMv2HelloStateRefresh,
    24: PIMv2HelloAddrList,
    None: _PIMv2GenericHello,
}


##################################
# PIMv2 Join/Prune
##################################
class PIMv2JoinPruneAddrsBase(_PIMGenericTlvBase):
    fields_desc = [
        ByteField("addr_family", 1),
        ByteField("encoding_type", 0),
        BitField("rsrvd", 0, 5),
        BitField("sparse", 0, 1),
        BitField("wildcard", 0, 1),
        BitField("rpt", 1, 1),
        ByteField("mask_len", 32),
        MultipleTypeField(
            [(IP6Field("src_ip", "::"),
              lambda pkt: pkt.addr_family == 2)],
            IPField("src_ip", "0.0.0.0")
        ),

    ]


class PIMv2JoinAddrs(PIMv2JoinPruneAddrsBase):
    name = "PIMv2 Join: Source Address"


class PIMv2PruneAddrs(PIMv2JoinPruneAddrsBase):
    name = "PIMv2 Prune: Source Address"


class PIMv2GroupAddrs(_PIMGenericTlvBase):
    name = "PIMv2 Join/Prune: Multicast Group Address"
    fields_desc = [
        ByteField("addr_family", 1),
        ByteField("encoding_type", 0),
        BitField("bidirection", 0, 1),
        BitField("reserved", 0, 6),
        BitField("admin_scope_zone", 0, 1),
        ByteField("mask_len", 32),
        MultipleTypeField(
            [(IP6Field("gaddr", "::"),
              lambda pkt: pkt.addr_family == 2)],
            IPField("gaddr", "0.0.0.0")
        ),
        BitFieldLenField("num_joins", None, size=16, count_of="join_ips"),
        BitFieldLenField("num_prunes", None, size=16, count_of="prune_ips"),
        PacketListField("join_ips", [], PIMv2JoinAddrs,
                        count_from=lambda x: x.num_joins),
        PacketListField("prune_ips", [], PIMv2PruneAddrs,
                        count_from=lambda x: x.num_prunes),
    ]


class PIMv2JoinPrune(_PIMGenericTlvBase):
    name = "PIMv2 Join/Prune Options"
    fields_desc = [
        ByteField("up_addr_family", 1),
        ByteField("up_encoding_type", 0),
        MultipleTypeField(
            [(IP6Field("up_neighbor_ip", "::"),
              lambda pkt: pkt.up_addr_family == 2)],
            IPField("up_neighbor_ip", "0.0.0.0")
        ),
        ByteField("reserved", 0),
        FieldLenField("num_group", None, count_of="jp_ips", fmt="B"),
        ShortField("holdtime", 210),
        PacketListField("jp_ips", [], PIMv2GroupAddrs,
                        count_from=lambda pkt: pkt.num_group)
    ]


bind_layers(IP, PIMv2Hdr, proto=103)
bind_layers(IPv6, PIMv2Hdr, nh=103)
bind_layers(PIMv2Hdr, PIMv2Hello, type=0)
bind_layers(PIMv2Hdr, PIMv2JoinPrune, type=3)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Andreas Korb <andreas.korb@e-mundo.de>
# Copyright (C) Friedrich Feigel <friedrich.feigel@e-mundo.de>
# Copyright (C) Nils Weiss <nils@we155.de>


import getopt
import sys
import signal
import re
import traceback

from ast import literal_eval

from scapy.config import conf
from scapy.consts import LINUX

if not LINUX or conf.use_pypy:
    conf.contribs['CANSocket'] = {'use-python-can': True}

from scapy.contrib.isotp import ISOTPSocket                    # noqa: E402
from scapy.contrib.cansocket import CANSocket, PYTHON_CAN      # noqa: E402
from scapy.contrib.automotive.obd.obd import OBD               # noqa: E402
from scapy.contrib.automotive.obd.scanner import OBD_Scanner, \
    OBD_S01_Enumerator, OBD_S02_Enumerator, OBD_S03_Enumerator, \
    OBD_S06_Enumerator, OBD_S07_Enumerator, OBD_S08_Enumerator, \
    OBD_S09_Enumerator, OBD_S0A_Enumerator  # noqa: E402


def signal_handler(sig, frame):
    print('Interrupting scan!')
    sys.exit(0)


def usage(is_error):
    print('''usage:\tobdscanner [-i|--interface] [-c|--channel] [-b|--bitrate]
                                [-a|--python-can_args] [-h|--help]
                                [-s|--source] [-d|--destination]
                                [-t|--timeout] [-f|--full]
                                [-v|--verbose]\n
    Scan for all possible obd service classes and their subfunctions.\n
    optional arguments:
    -c, --channel               python-can channel or Linux SocketCAN interface name\n
    additional required arguments for WINDOWS or Python 2:
    -i, --interface             python-can interface for the scan.
                                Depends on used interpreter and system,
                                see examples below. Any python-can interface can
                                be provided. Please see:
                                https://python-can.readthedocs.io for
                                further interface examples.
    optional arguments:
    -a, --python-can_args       Additional arguments for a python-can Bus object.
    -h, --help                  show this help message and exit
    -s, --source                ISOTP-socket source id (hex)
    -d, --destination           ISOTP-socket destination id (hex)
    -t, --timeout               Timeout after which the scanner proceeds to next service [seconds]
    -f, --full                  Full scan on id services
    -v, --verbose               Display information during scan
    -1                          Scan OBD Service 01
    -2                          Scan OBD Service 02
    -3                          Scan OBD Service 03
    -6                          Scan OBD Service 06
    -7                          Scan OBD Service 07
    -8                          Scan OBD Service 08
    -9                          Scan OBD Service 09
    -A                          Scan OBD Service 0A\n
    Example of use:\n
    Python2 or Windows:
    python2 -m scapy.tools.automotive.obdscanner --interface=pcan --channel=PCAN_USBBUS1 --source=0x070 --destination 0x034
    python2 -m scapy.tools.automotive.obdscanner --interface vector --channel 0 --source 0x000 --destination 0x734
    python2 -m scapy.tools.automotive.obdscanner --interface socketcan --channel=can0 --source 0x089 --destination 0x234
    python2 -m scapy.tools.automotive.obdscanner --interface vector --channel 0 --python-can_args 'bitrate=500000, poll_interval=1' --source=0x070 --destination 0x034\n
    Python3 on Linux:
    python3 -m scapy.tools.automotive.obdscanner --channel can0 --source 0x123 --destination 0x456 \n''',  # noqa: E501
          file=sys.stderr if is_error else sys.stdout)


def get_can_socket(channel, interface, python_can_args):
    if PYTHON_CAN:
        if python_can_args:
            arg_dict = dict((k, literal_eval(v)) for k, v in
                            (pair.split('=') for pair in
                             re.split(', | |,', python_can_args)))
            return CANSocket(bustype=interface, channel=channel, **arg_dict)
        else:
            return CANSocket(bustype=interface, channel=channel)
    else:
        return CANSocket(channel=channel)


def get_isotp_socket(csock, source, destination):
    return ISOTPSocket(csock, source, destination, basecls=OBD, padding=True)


def run_scan(isock, enumerators, full_scan, verbose, timeout):
    s = OBD_Scanner(isock, test_cases=enumerators, full_scan=full_scan,
                    debug=verbose,
                    timeout=timeout)
    print("Starting OBD-Scan...")
    s.scan()
    s.show_testcases()


def main():

    channel = None
    interface = None
    source = 0x7e0
    destination = 0x7df
    timeout = 0.1
    full_scan = False
    verbose = False
    python_can_args = None
    enumerators = []
    conf.verb = -1

    options = getopt.getopt(
        sys.argv[1:],
        'i:c:s:d:a:t:hfv1236789A',
        ['interface=', 'channel=', 'source=', 'destination=',
         'help', 'timeout=', 'python-can_args=', 'full',
         'verbose'])

    try:
        for opt, arg in options[0]:
            if opt in ('-i', '--interface'):
                interface = arg
            elif opt in ('-c', '--channel'):
                channel = arg
            elif opt in ('-a', '--python-can_args'):
                python_can_args = arg
            elif opt in ('-s', '--source'):
                source = int(arg, 16)
            elif opt in ('-d', '--destination'):
                destination = int(arg, 16)
            elif opt in ('-h', '--help'):
                usage(False)
                sys.exit(0)
            elif opt in ('-t', '--timeout'):
                timeout = float(arg)
            elif opt in ('-f', '--full'):
                full_scan = True
            elif opt == '-1':
                enumerators += [OBD_S01_Enumerator]
            elif opt == '-2':
                enumerators += [OBD_S02_Enumerator]
            elif opt == '-3':
                enumerators += [OBD_S03_Enumerator]
            elif opt == '-6':
                enumerators += [OBD_S06_Enumerator]
            elif opt == '-7':
                enumerators += [OBD_S07_Enumerator]
            elif opt == '-8':
                enumerators += [OBD_S08_Enumerator]
            elif opt == '-9':
                enumerators += [OBD_S09_Enumerator]
            elif opt == '-A':
                enumerators += [OBD_S0A_Enumerator]
            elif opt in ('-v', '--verbose'):
                verbose = True
    except getopt.GetoptError as msg:
        usage(True)
        print("ERROR:", msg, file=sys.stderr)
        raise SystemExit

    if channel is None or \
            (PYTHON_CAN and interface is None):
        usage(True)
        print("\nPlease provide all required arguments.\n",
              file=sys.stderr)
        sys.exit(1)

    if 0 > source >= 0x800 or 0 > destination >= 0x800\
            or source == destination:
        print("The ids must be >= 0 and < 0x800 and not equal.",
              file=sys.stderr)
        sys.exit(1)

    if 0 > timeout:
        print("The timeout must be a positive value")
        sys.exit(1)

    csock = None
    isock = None
    try:
        csock = get_can_socket(channel, interface, python_can_args)
        isock = get_isotp_socket(csock, source, destination)

        signal.signal(signal.SIGINT, signal_handler)
        run_scan(isock, enumerators, full_scan, verbose, timeout)

    except Exception as e:
        usage(True)
        print("\nSocket couldn't be created. Check your arguments.\n",
              file=sys.stderr)
        print(e, file=sys.stderr)
        if verbose:
            traceback.print_exc(file=sys.stderr)
        sys.exit(1)

    finally:
        if isock:
            isock.close()
        if csock:
            csock.close()


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2007, 2008, 2009 Arnaud Ebalard
#                     2015, 2016 Maxence Tury

"""
Cryptographic capabilities for TLS.
"""
